from __future__ import annotations

import json
import logging
import uuid
from datetime import datetime, timedelta, timezone
from enum import Enum
from typing import Any

import jwt

from .api_common import ApiCommon
from .const import HEADER_ACCESS_TOKEN, HEADER_AUGUST_ACCESS_TOKEN
from .time import parse_datetime

# The default time before expiration to refresh a token
DEFAULT_RENEWAL_THRESHOLD = timedelta(days=7)

_LOGGER = logging.getLogger(__name__)


def to_authentication_json(authentication):
    if authentication is None:
        return json.dumps({})

    return json.dumps(
        {
            "install_id": authentication.install_id,
            "access_token": authentication.access_token,
            "access_token_expires": authentication.access_token_expires,
            "state": authentication.state.value,
        }
    )


def from_authentication_json(data):
    if data is None:
        return None

    install_id = data["install_id"]
    access_token = data["access_token"]
    access_token_expires = data["access_token_expires"]
    state = AuthenticationState(data["state"])
    return Authentication(state, install_id, access_token, access_token_expires)


class Authentication:
    def __init__(
        self, state, install_id=None, access_token=None, access_token_expires=None
    ):
        self._state = state
        self._install_id = str(uuid.uuid4()) if install_id is None else install_id
        self._access_token = access_token
        self._access_token_expires = access_token_expires
        self._parsed_expiration_time = None
        if access_token_expires:
            self._parsed_expiration_time = parse_datetime(access_token_expires)

    @property
    def install_id(self):
        return self._install_id

    @property
    def access_token(self):
        return self._access_token

    @property
    def access_token_expires(self):
        return self._access_token_expires

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, value):
        self._state = value

    def parsed_expiration_time(self):
        return self._parsed_expiration_time

    def is_expired(self):
        return self._parsed_expiration_time < datetime.now(timezone.utc)


class AuthenticationState(Enum):
    REQUIRES_AUTHENTICATION = "requires_authentication"
    REQUIRES_VALIDATION = "requires_validation"
    AUTHENTICATED = "authenticated"
    BAD_PASSWORD = "bad_password"  # nosec


class ValidationResult(Enum):
    VALIDATED = "validated"
    INVALID_VERIFICATION_CODE = "invalid_verification_code"


class AuthenticatorCommon:
    def __init__(
        self,
        api: ApiCommon,
        login_method: str | None,
        username: str | None,
        password: str | None,
        install_id: str | None = None,
        access_token_cache_file: str | None = None,
        access_token_renewal_threshold: timedelta = DEFAULT_RENEWAL_THRESHOLD,
    ) -> None:
        self._api = api
        self._login_method = login_method
        self._username = username
        self._password = password
        self._install_id = install_id
        self._access_token_cache_file = access_token_cache_file
        self._access_token_renewal_threshold = access_token_renewal_threshold
        self._authentication = None

    def _authentication_from_session_response(
        self,
        install_id: str,
        response_headers: dict[str, Any],
        json_dict: dict[str, Any],
    ) -> Authentication:
        access_token = (
            response_headers.get(HEADER_ACCESS_TOKEN)
            or response_headers[HEADER_AUGUST_ACCESS_TOKEN]
        )
        access_token_expires = json_dict["expiresAt"]
        v_password = json_dict["vPassword"]
        v_install_id = json_dict["vInstallId"]

        if not v_password:
            state = AuthenticationState.BAD_PASSWORD
        elif not v_install_id:
            state = AuthenticationState.REQUIRES_VALIDATION
        else:
            state = AuthenticationState.AUTHENTICATED

        self._authentication = Authentication(
            state, install_id, access_token, access_token_expires
        )

        return self._authentication

    def should_refresh(self):
        return self._authentication.state == AuthenticationState.AUTHENTICATED and (
            (self._authentication.parsed_expiration_time() - datetime.now(timezone.utc))
            < self._access_token_renewal_threshold
        )

    def _process_refreshed_access_token(self, refreshed_token):
        jwt_claims = jwt.decode(refreshed_token, options={"verify_signature": False})

        if "exp" not in jwt_claims:
            _LOGGER.warning("Did not find expected `exp' claim in JWT")
            return self._authentication

        new_expiration = datetime.utcfromtimestamp(jwt_claims["exp"])  # noqa: DTZ004
        # The yale access api always returns expiresAt in the format
        # '%Y-%m-%dT%H:%M:%S.%fZ'
        # from the get_session api call
        # It is important we store access_token_expires formatted
        # the same way for compatibility
        self._authentication = Authentication(
            self._authentication.state,
            install_id=self._authentication.install_id,
            access_token=refreshed_token,
            access_token_expires=new_expiration.strftime("%Y-%m-%dT%H:%M:%S.%fZ"),
        )

        _LOGGER.info("Successfully refreshed access token")
        return self._authentication



from __future__ import annotations

import logging
from typing import Any

from ._compat import cached_property
from .backports.enum import StrEnum
from .device import Device, DeviceDetail


class ArmState(StrEnum):
    Away = "FULL_ARM"
    Home = "PARTIAL_ARM"
    Disarm = "DISARM"


_LOGGER = logging.getLogger(__name__)


class Alarm(Device):
    """Class to hold details about an alarm."""

    def __init__(self, device_id: str, data: dict[str, Any]) -> None:
        _LOGGER.info("Alarm init - %s", data["location"])
        super().__init__(device_id, data["location"], data["houseID"])
        self._pubsub_channel = data["pubsubChannel"]
        self._serial_number = data["serialNumber"]
        self._status = data["status"]
        self._areaIDs = data["areaIDs"]

    @cached_property
    def pubsub_channel(self):
        return self._pubsub_channel

    @cached_property
    def serial_number(self):
        return self._serial_number

    @cached_property
    def status(self):
        return self._status

    @cached_property
    def areaIDs(self):
        return self._areaIDs

    def __repr__(self):
        return f"Alarm(id={self.device_id}, name={self.device_name}, house_id={self.house_id})"


class AlarmDevice(DeviceDetail):
    """Class to hold details about a device attached to the alarm."""

    def __init__(self, data: dict[str, Any]) -> None:
        _LOGGER.info("Alarm init - %s (%s)", data["name"], data["type"])
        super().__init__(
            data["_id"],
            data["name"],
            data["alarmID"],
            data["serialNumber"],
            data["status"]["firmwareVersion"],
            data.get("pubsubChannel"),
            data,
        )

        self._status: str = data["status"]
        self._model = data["type"]

        self._battery_level = 100
        if self._status.get("lowBattery", False):
            self._battery_level = 10

    @cached_property
    def status(self) -> str:
        return self._status

    @cached_property
    def model(self) -> str:
        return self._model

    @cached_property
    def is_online(self) -> bool:
        return self.status.get("online", False)

    @cached_property
    def contact_open(self) -> bool:
        return self.status.get("contactOpen", False)

    @cached_property
    def fault(self) -> bool:
        return self.status.get("fault", False)

    @cached_property
    def tamperOpen(self) -> bool:
        return self.status.get("tamperOpen", False)

    @cached_property
    def battery_level(self) -> int | None:
        """Return an approximation of the battery percentage."""
        return self._battery_level

    def __repr__(self):
        return f"AlarmDevice(id={self.device_id}, name={self.device_name}, type={self.model}, alarm_id={self.house_id})"



from __future__ import annotations

from .time import parse_datetime


class Pin:
    def __init__(self, data):
        self._pin_id = data["_id"]
        self._lock_id = data["lockID"]
        self._user_id = data["userID"]
        self._state = data["state"]
        self._pin = data["pin"]
        self._slot = data["slot"]
        self._access_type = data["accessType"]
        self._first_name = data["firstName"]
        self._last_name = data["lastName"]
        self._unverified = data["unverified"]

        self._created_at = data["createdAt"]
        self._updated_at = data["updatedAt"]
        self._loaded_date = data["loadedDate"]
        # Times for temporary access codes
        self._access_start_time = data.get("accessStartTime")
        self._access_end_time = data.get("accessEndTime")
        self._access_times = data.get("accessTimes")

    @property
    def pin_id(self):
        return self._pin_id

    @property
    def lock_id(self):
        return self._lock_id

    @property
    def user_id(self):
        return self._user_id

    @property
    def state(self):
        return self._state

    @property
    def pin(self):
        return self._pin

    @property
    def slot(self):
        return self._slot

    @property
    def access_type(self):
        return self._access_type

    @property
    def first_name(self):
        return self._first_name

    @property
    def last_name(self):
        return self._last_name

    @property
    def unverified(self):
        return self._unverified

    @property
    def created_at(self):
        return parse_datetime(self._created_at)

    @property
    def updated_at(self):
        return parse_datetime(self._updated_at)

    @property
    def loaded_date(self):
        return parse_datetime(self._loaded_date)

    @property
    def access_start_time(self):
        if not self._access_start_time:
            return None
        return parse_datetime(self._access_start_time)

    @property
    def access_end_time(self):
        if not self._access_end_time:
            return None
        return parse_datetime(self._access_end_time)

    @property
    def access_times(self):
        if not self._access_times:
            return None
        return parse_datetime(self._access_times)

    def __repr__(self):
        return f"Pin(id={self.pin_id} firstName={self.first_name}, lastName={self.last_name})"



import json
from datetime import datetime, timedelta, timezone

import aiounittest
from aiohttp import ClientError, ClientSession
from aioresponses import aioresponses
from dateutil.tz import tzutc

from yalexs.api_async import ApiAsync
from yalexs.api_common import (
    API_GET_HOUSES_URL,
    API_GET_SESSION_URL,
    API_SEND_VERIFICATION_CODE_URLS,
    API_VALIDATE_VERIFICATION_CODE_URLS,
    ApiCommon,
)
from yalexs.authenticator_async import (
    AuthenticationState,
    AuthenticatorAsync,
    ValidationResult,
)
from yalexs.const import DEFAULT_BRAND, HEADER_AUGUST_ACCESS_TOKEN


def format_datetime(dt):
    return dt.strftime("%Y-%m-%d %H:%M:%S.%f")[:-3] + "Z"


class TestAuthenticatorAsync(aiounittest.AsyncTestCase):
    def setUp(self):
        """Setup things to be run when tests are started."""

    async def _async_create_authenticator_async(self, mock_aioresponses):
        authenticator = AuthenticatorAsync(
            ApiAsync(ClientSession()), "phone", "user", "pass", install_id="install_id"
        )
        await authenticator.async_setup_authentication()
        return authenticator

    def _setup_session_response(
        self,
        mock_aioresponses,
        v_password,
        v_install_id,
        expires_at=format_datetime(datetime.utcnow()),  # noqa: DTZ003
    ):
        mock_aioresponses.post(
            ApiCommon(DEFAULT_BRAND).get_brand_url(API_GET_SESSION_URL),
            headers={"x-august-access-token": "access_token"},
            body=json.dumps(
                {
                    "expiresAt": expires_at,
                    "vPassword": v_password,
                    "vInstallId": v_install_id,
                }
            ),
        )

    @aioresponses()
    async def test_async_should_refresh_when_token_expiry_is_after_renewal_threshold(
        self, mock_aioresponses
    ):
        expired_expires_at = format_datetime(
            datetime.now(timezone.utc) + timedelta(days=6)
        )
        self._setup_session_response(
            mock_aioresponses, True, True, expires_at=expired_expires_at
        )

        authenticator = await self._async_create_authenticator_async(mock_aioresponses)
        await authenticator.async_authenticate()

        should_refresh = authenticator.should_refresh()

        self.assertEqual(True, should_refresh)

    @aioresponses()
    async def test_async_should_refresh_when_token_expiry_is_before_renewal_threshold(
        self, mock_aioresponses
    ):
        not_expired_expires_at = format_datetime(
            datetime.now(timezone.utc) + timedelta(days=8)
        )
        self._setup_session_response(
            mock_aioresponses, True, True, expires_at=not_expired_expires_at
        )

        authenticator = await self._async_create_authenticator_async(mock_aioresponses)
        await authenticator.async_authenticate()

        should_refresh = authenticator.should_refresh()

        self.assertEqual(False, should_refresh)

    @aioresponses()
    async def test_async_refresh_token(self, mock_aioresponses):
        self._setup_session_response(mock_aioresponses, True, True)

        authenticator = await self._async_create_authenticator_async(mock_aioresponses)
        await authenticator.async_authenticate()

        token = "e30=.eyJleHAiOjEzMzd9.e30="
        mock_aioresponses.get(
            ApiCommon(DEFAULT_BRAND).get_brand_url(API_GET_HOUSES_URL),
            body=token,
            headers={HEADER_AUGUST_ACCESS_TOKEN: token},
        )

        access_token = await authenticator.async_refresh_access_token(force=False)

        self.assertEqual(token, access_token.access_token)
        self.assertEqual(
            datetime.fromtimestamp(1337, tz=tzutc()),
            access_token.parsed_expiration_time(),
        )

    @aioresponses()
    async def test_async_get_session_with_authenticated_response(
        self, mock_aioresponses
    ):
        self._setup_session_response(mock_aioresponses, True, True)

        authenticator = await self._async_create_authenticator_async(mock_aioresponses)
        authentication = await authenticator.async_authenticate()

        self.assertEqual("access_token", authentication.access_token)
        self.assertEqual("install_id", authentication.install_id)
        self.assertEqual(AuthenticationState.AUTHENTICATED, authentication.state)

    @aioresponses()
    async def test_async_get_session_with_bad_password_response(
        self, mock_aioresponses
    ):
        self._setup_session_response(mock_aioresponses, False, True)

        authenticator = await self._async_create_authenticator_async(mock_aioresponses)
        authentication = await authenticator.async_authenticate()

        self.assertEqual("access_token", authentication.access_token)
        self.assertEqual("install_id", authentication.install_id)
        self.assertEqual(AuthenticationState.BAD_PASSWORD, authentication.state)

    @aioresponses()
    async def test_async_get_session_with_requires_validation_response(
        self, mock_aioresponses
    ):
        self._setup_session_response(mock_aioresponses, True, False)

        authenticator = await self._async_create_authenticator_async(mock_aioresponses)
        authentication = await authenticator.async_authenticate()

        self.assertEqual("access_token", authentication.access_token)
        self.assertEqual("install_id", authentication.install_id)
        self.assertEqual(AuthenticationState.REQUIRES_VALIDATION, authentication.state)

    @aioresponses()
    async def test_async_get_session_with_already_authenticated_state(
        self, mock_aioresponses
    ):
        self._setup_session_response(mock_aioresponses, True, True)

        authenticator = await self._async_create_authenticator_async(mock_aioresponses)
        # this will set authentication state to AUTHENTICATED
        await authenticator.async_authenticate()
        # call authenticate() again
        authentication = await authenticator.async_authenticate()

        self.assertEqual("access_token", authentication.access_token)
        self.assertEqual("install_id", authentication.install_id)
        self.assertEqual(AuthenticationState.AUTHENTICATED, authentication.state)

    @aioresponses()
    async def test_async_send_verification_code(self, mock_aioresponses):
        self._setup_session_response(mock_aioresponses, True, False)

        authenticator = await self._async_create_authenticator_async(mock_aioresponses)
        mock_aioresponses.post(
            ApiCommon(DEFAULT_BRAND).get_brand_url(
                API_SEND_VERIFICATION_CODE_URLS["phone"]
            ),
            body="{}",
        )
        await authenticator.async_authenticate()
        result = await authenticator.async_send_verification_code()

        self.assertEqual(True, result)

    @aioresponses()
    async def test_async_validate_verification_code_with_no_code(
        self, mock_aioresponses
    ):
        self._setup_session_response(mock_aioresponses, True, False)

        authenticator = await self._async_create_authenticator_async(mock_aioresponses)
        await authenticator.async_authenticate()

        mock_aioresponses.post(
            ApiCommon(DEFAULT_BRAND).get_brand_url(
                API_VALIDATE_VERIFICATION_CODE_URLS["phone"]
            ),
            body="{}",
        )
        result = await authenticator.async_validate_verification_code("")

        # mock_aioresponses.async_validate_verification_code.assert_not_called()

        self.assertEqual(ValidationResult.INVALID_VERIFICATION_CODE, result)

    @aioresponses()
    async def test_async_validate_verification_code_with_validated_response(
        self, mock_aioresponses
    ):
        self._setup_session_response(mock_aioresponses, True, False)

        mock_aioresponses.post(
            ApiCommon(DEFAULT_BRAND).get_brand_url(
                API_VALIDATE_VERIFICATION_CODE_URLS["phone"]
            ),
            body="{}",
        )

        authenticator = await self._async_create_authenticator_async(mock_aioresponses)
        await authenticator.async_authenticate()
        result = await authenticator.async_validate_verification_code("123456")

        self.assertEqual(ValidationResult.VALIDATED, result)

    @aioresponses()
    async def test_async_validate_verification_code_with_invalid_code_response(
        self, mock_aioresponses
    ):
        self._setup_session_response(mock_aioresponses, True, False)

        mock_aioresponses.post(
            ApiCommon(DEFAULT_BRAND).get_brand_url(
                API_VALIDATE_VERIFICATION_CODE_URLS["phone"]
            ),
            exception=ClientError(),
        )

        authenticator = await self._async_create_authenticator_async(mock_aioresponses)
        await authenticator.async_authenticate()
        result = await authenticator.async_validate_verification_code("123456")

        self.assertEqual(ValidationResult.INVALID_VERIFICATION_CODE, result)



from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import Any, Union

from ._compat import cached_property
from .backports.enum import StrEnum
from .lock import LockDoorStatus, LockStatus
from .time import epoch_to_datetime, parse_datetime
from .users import YaleUser, get_user_info

ACTION_LOCK_ONETOUCHLOCK = "onetouchlock"
ACTION_LOCK_ONETOUCHLOCK_2 = "one_touch_lock"
ACTION_LOCK_LOCK = "lock"
ACTION_RF_LOCK = "rf_lock"
ACTION_RF_SECURE = "rf_secure"
ACTION_RF_UNLATCH = "rf_unlatch"
ACTION_RF_UNLOCK = "rf_unlock"
ACTION_LOCK_AUTO_LOCK = "auto_lock"
ACTION_LOCK_BLE_LOCK = "ble_lock"
ACTION_LOCK_BLE_UNLATCH = "ble_unlatch"
ACTION_LOCK_BLE_UNLOCK = "ble_unlock"
ACTION_LOCK_REMOTE_LOCK = "remote_lock"
ACTION_LOCK_REMOTE_UNLATCH = "remote_unlatch"
ACTION_LOCK_REMOTE_UNLOCK = "remote_unlock"
ACTION_LOCK_PIN_LOCK = "pin_lock"
ACTION_LOCK_PIN_UNLATCH = "pin_unlatch"
ACTION_LOCK_PIN_UNLOCK = "pin_unlock"
ACTION_LOCK_MANUAL_LOCK = "manual_lock"
ACTION_LOCK_MANUAL_UNLATCH = "manual_unlatch"
ACTION_LOCK_MANUAL_UNLOCK = "manual_unlock"
ACTION_LOCK_LOCKING = "locking"
ACTION_LOCK_UNLATCH = "unlatch"
ACTION_LOCK_UNLATCHING = "unlatching"
ACTION_LOCK_UNLOCK = "unlock"
ACTION_LOCK_UNLOCKING = "unlocking"
ACTION_LOCK_JAMMED = "jammed"
ACTION_HOMEKEY_LOCK = "homekey_lock"
ACTION_HOMEKEY_UNLATCH = "homekey_unlatch"
ACTION_HOMEKEY_UNLOCK = "homekey_unlock"
ACTION_LINKED_LOCK = "linked_lock"
ACTION_LINKED_UNLOCK = "linked_unlock"
ACTION_LOCK_FINGER_LOCK = "finger_lock"
ACTION_LOCK_FINGER_UNLOCK = "finger_unlock"

ACTION_DOOR_OPEN = "dooropen"
ACTION_DOOR_OPEN_2 = "door_open"
ACTION_DOOR_CLOSED = "doorclosed"
ACTION_DOOR_CLOSE_2 = "door_close"


ACTION_DOORBELL_CALL_INITIATED = "doorbell_call_initiated"
ACTION_DOORBELL_MOTION_DETECTED = "doorbell_motion_detected"
ACTION_DOORBELL_CALL_MISSED = "doorbell_call_missed"
ACTION_DOORBELL_CALL_HANGUP = "doorbell_call_hangup"
ACTION_LOCK_DOORBELL_BUTTON_PUSHED = "lock_accessory_motion_detect"

ACTION_BRIDGE_ONLINE = "associated_bridge_online"  # pubnub only
ACTION_BRIDGE_OFFLINE = "associated_bridge_offline"  # pubnub only
ACTION_DOORBELL_IMAGE_CAPTURE = "imagecapture"  # pubnub only
ACTION_DOORBELL_BUTTON_PUSHED = "buttonpush"  # pubnub only

ACTIVITY_ACTIONS_BRIDGE_OPERATION = {ACTION_BRIDGE_ONLINE, ACTION_BRIDGE_OFFLINE}

ACTIVITY_ACTIONS_DOORBELL_DING = {
    ACTION_DOORBELL_BUTTON_PUSHED,
    ACTION_DOORBELL_CALL_MISSED,
    ACTION_DOORBELL_CALL_HANGUP,
    ACTION_LOCK_DOORBELL_BUTTON_PUSHED,
}
ACTIVITY_ACTIONS_DOORBELL_IMAGE_CAPTURE = {ACTION_DOORBELL_IMAGE_CAPTURE}
ACTIVITY_ACTIONS_DOORBELL_MOTION = {ACTION_DOORBELL_MOTION_DETECTED}
ACTIVITY_ACTIONS_DOORBELL_VIEW = {ACTION_DOORBELL_CALL_INITIATED}
ACTIVITY_ACTIONS_LOCK_OPERATION = {
    ACTION_RF_SECURE,
    ACTION_RF_LOCK,
    ACTION_RF_UNLATCH,
    ACTION_RF_UNLOCK,
    ACTION_HOMEKEY_LOCK,
    ACTION_HOMEKEY_UNLATCH,
    ACTION_HOMEKEY_UNLOCK,
    ACTION_LINKED_LOCK,
    ACTION_LINKED_UNLOCK,
    ACTION_LOCK_AUTO_LOCK,
    ACTION_LOCK_ONETOUCHLOCK,
    ACTION_LOCK_ONETOUCHLOCK_2,
    ACTION_LOCK_LOCK,
    ACTION_LOCK_UNLATCH,
    ACTION_LOCK_UNLOCK,
    ACTION_LOCK_LOCKING,
    ACTION_LOCK_UNLATCHING,
    ACTION_LOCK_UNLOCKING,
    ACTION_LOCK_JAMMED,
    ACTION_LOCK_BLE_LOCK,
    ACTION_LOCK_BLE_UNLATCH,
    ACTION_LOCK_BLE_UNLOCK,
    ACTION_LOCK_REMOTE_LOCK,
    ACTION_LOCK_REMOTE_UNLATCH,
    ACTION_LOCK_REMOTE_UNLOCK,
    ACTION_LOCK_PIN_LOCK,
    ACTION_LOCK_PIN_UNLATCH,
    ACTION_LOCK_PIN_UNLOCK,
    ACTION_LOCK_MANUAL_LOCK,
    ACTION_LOCK_MANUAL_UNLATCH,
    ACTION_LOCK_MANUAL_UNLOCK,
    ACTION_LOCK_FINGER_LOCK,
    ACTION_LOCK_FINGER_UNLOCK,
}


ACTIVITY_TO_FIRST_LAST_NAME = {
    ACTION_RF_SECURE: ("Radio Frequency", "Secure"),
    ACTION_RF_LOCK: ("Radio Frequency", "Lock"),
    ACTION_RF_UNLATCH: ("Radio Frequency", "Unlatch"),
    ACTION_RF_UNLOCK: ("Radio Frequency", "Unlock"),
    ACTION_HOMEKEY_LOCK: ("Homekey", "Lock"),
    ACTION_HOMEKEY_UNLATCH: ("Homekey", "Unlatch"),
    ACTION_HOMEKEY_UNLOCK: ("Homekey", "Unlock"),
    ACTION_LINKED_LOCK: ("Linked", "Lock"),
    ACTION_LINKED_UNLOCK: ("Linked", "Unlock"),
    ACTION_LOCK_AUTO_LOCK: ("Auto", "Lock"),
    ACTION_LOCK_ONETOUCHLOCK: ("One-Touch", "Lock"),
    ACTION_LOCK_ONETOUCHLOCK_2: ("One-Touch", "Lock"),
    ACTION_LOCK_BLE_LOCK: ("Bluetooth", "Lock"),
    ACTION_LOCK_BLE_UNLATCH: ("Bluetooth", "Unlatch"),
    ACTION_LOCK_BLE_UNLOCK: ("Bluetooth", "Unlock"),
    ACTION_LOCK_MANUAL_LOCK: ("Manual", "Lock"),
    ACTION_LOCK_MANUAL_UNLATCH: ("Manual", "Unlatch"),
    ACTION_LOCK_MANUAL_UNLOCK: ("Manual", "Unlock"),
    ACTION_LOCK_FINGER_LOCK: ("Fingerprint", "Lock"),
    ACTION_LOCK_FINGER_UNLOCK: ("Fingerprint", "Unlock"),
    ACTION_LOCK_PIN_LOCK: ("Pin", "Lock"),
    ACTION_LOCK_PIN_UNLATCH: ("Pin", "Unlatch"),
    ACTION_LOCK_PIN_UNLOCK: ("Pin", "Unlock"),
}


ACTIVITY_ACTIONS_DOOR_OPERATION = {
    ACTION_DOOR_CLOSED,
    ACTION_DOOR_OPEN,
    ACTION_DOOR_OPEN_2,
    ACTION_DOOR_CLOSE_2,
}

KEYPAD_ACTIONS = {
    ACTION_LOCK_ONETOUCHLOCK,
    ACTION_LOCK_ONETOUCHLOCK_2,
    ACTION_LOCK_PIN_LOCK,
    ACTION_LOCK_PIN_UNLATCH,
    ACTION_LOCK_PIN_UNLOCK,
    ACTION_LOCK_FINGER_LOCK,
    ACTION_LOCK_FINGER_UNLOCK,
}
REMOTE_ACTIONS = {
    ACTION_LOCK_REMOTE_LOCK,
    ACTION_LOCK_REMOTE_UNLATCH,
    ACTION_LOCK_REMOTE_UNLOCK,
}
AUTO_RELOCK_ACTIONS = {ACTION_LOCK_AUTO_LOCK}

TAG_ACTIONS = {
    ACTION_RF_SECURE,
    ACTION_RF_LOCK,
    ACTION_RF_UNLATCH,
    ACTION_RF_UNLOCK,
    ACTION_HOMEKEY_LOCK,
    ACTION_HOMEKEY_UNLATCH,
    ACTION_HOMEKEY_UNLOCK,
}

MANUAL_ACTIONS = {
    ACTION_LOCK_MANUAL_LOCK,
    ACTION_LOCK_MANUAL_UNLATCH,
    ACTION_LOCK_MANUAL_UNLOCK,
}

ACTIVITY_ACTION_STATES: dict[str, LockStatus | LockDoorStatus] = {
    ACTION_RF_SECURE: LockStatus.LOCKED,
    ACTION_RF_LOCK: LockStatus.LOCKED,
    ACTION_RF_UNLATCH: LockStatus.UNLATCHED,
    ACTION_RF_UNLOCK: LockStatus.UNLOCKED,
    ACTION_HOMEKEY_LOCK: LockStatus.LOCKED,
    ACTION_LINKED_LOCK: LockStatus.LOCKED,
    ACTION_LINKED_UNLOCK: LockStatus.UNLOCKED,
    ACTION_HOMEKEY_UNLATCH: LockStatus.UNLATCHED,
    ACTION_HOMEKEY_UNLOCK: LockStatus.UNLOCKED,
    ACTION_LOCK_AUTO_LOCK: LockStatus.LOCKED,
    ACTION_LOCK_ONETOUCHLOCK: LockStatus.LOCKED,
    ACTION_LOCK_ONETOUCHLOCK_2: LockStatus.LOCKED,
    ACTION_LOCK_LOCK: LockStatus.LOCKED,
    ACTION_LOCK_UNLATCH: LockStatus.UNLATCHED,
    ACTION_LOCK_UNLOCK: LockStatus.UNLOCKED,
    ACTION_LOCK_LOCKING: LockStatus.LOCKING,
    ACTION_LOCK_UNLATCHING: LockStatus.UNLATCHING,
    ACTION_LOCK_UNLOCKING: LockStatus.UNLOCKING,
    ACTION_LOCK_JAMMED: LockStatus.JAMMED,
    ACTION_DOOR_OPEN: LockDoorStatus.OPEN,
    ACTION_DOOR_CLOSED: LockDoorStatus.CLOSED,
    ACTION_DOOR_OPEN_2: LockDoorStatus.OPEN,
    ACTION_DOOR_CLOSE_2: LockDoorStatus.CLOSED,
    ACTION_LOCK_BLE_LOCK: LockStatus.LOCKED,
    ACTION_LOCK_BLE_UNLATCH: LockStatus.UNLATCHED,
    ACTION_LOCK_BLE_UNLOCK: LockStatus.UNLOCKED,
    ACTION_LOCK_REMOTE_LOCK: LockStatus.LOCKED,
    ACTION_LOCK_REMOTE_UNLATCH: LockStatus.UNLATCHED,
    ACTION_LOCK_REMOTE_UNLOCK: LockStatus.UNLOCKED,
    ACTION_LOCK_PIN_LOCK: LockStatus.LOCKED,
    ACTION_LOCK_PIN_UNLATCH: LockStatus.UNLATCHED,
    ACTION_LOCK_PIN_UNLOCK: LockStatus.UNLOCKED,
    ACTION_LOCK_MANUAL_LOCK: LockStatus.LOCKED,
    ACTION_LOCK_MANUAL_UNLATCH: LockStatus.UNLATCHED,
    ACTION_LOCK_MANUAL_UNLOCK: LockStatus.UNLOCKED,
    ACTION_LOCK_FINGER_LOCK: LockStatus.LOCKED,
    ACTION_LOCK_FINGER_UNLOCK: LockStatus.UNLOCKED,
}


class Source(StrEnum):
    """Source of the activity."""

    LOCK_OPERATE = "lock_operate"
    PUBNUB = "pubnub"
    LOG = "log"
    WEBSOCKET = "websocket"


SOURCE_LOCK_OPERATE = Source.LOCK_OPERATE
SOURCE_PUBNUB = Source.PUBNUB
SOURCE_LOG = Source.LOG
SOURCE_WEBSOCKET = Source.WEBSOCKET

# If we get a lock operation activity with the same time stamp as a moving
# activity we want to use the non-moving activity since its the completed state.
MOVING_STATES: set[LockStatus | LockDoorStatus] = {
    LockStatus.UNLOCKING,
    LockStatus.UNLATCHING,
    LockStatus.LOCKING,
}

ACTIVITY_MOVING_STATES = {
    action
    for action, action_state in ACTIVITY_ACTION_STATES.items()
    if action_state in MOVING_STATES
}


class ActivityType(Enum):
    DOORBELL_MOTION = "doorbell_motion"
    DOORBELL_DING = "doorbell_ding"
    DOORBELL_VIEW = "doorbell_view"
    LOCK_OPERATION = "lock_operation"
    LOCK_OPERATION_WITHOUT_OPERATOR = "lock_operation_without_operator"
    DOOR_OPERATION = "door_operation"
    BRIDGE_OPERATION = "bridge_operation"
    DOORBELL_IMAGE_CAPTURE = "doorbell_image_capture"


class Activity:
    """Base class for activities."""

    def __init__(self, source: str, data: dict[str, Any]) -> None:
        """Initialize activity."""
        self._source = source
        self._data = data

    def __repr__(self):
        """Return the representation."""
        return (
            f"<{self.__class__.__name__} action={self.action} activity_type={self.activity_type} "
            f"activity_start_time={self.activity_start_time} "
            f"device_name={self.device_name}>"
        )

    @cached_property
    def _entities(self) -> dict[str, Any]:
        """Return the entities of the activity."""
        return self._data.get("entities", {})

    @cached_property
    def _info(self) -> dict[str, Any]:
        """Return the info of the activity."""
        return self._data.get("info", {})

    @cached_property
    def was_pushed(self) -> bool:
        """Return if the activity was pushed."""
        return self._source in (SOURCE_PUBNUB, SOURCE_WEBSOCKET)

    @cached_property
    def source(self) -> str:
        """Return the source of the activity."""
        return self._source

    @cached_property
    def activity_type(self) -> ActivityType:
        """Return the type of the activity."""
        return self._activity_type

    @cached_property
    def activity_id(self) -> str | None:
        """Return the ID of the activity."""
        return self._entities.get("activity")

    @cached_property
    def house_id(self) -> str | None:
        """Return the house ID of the activity."""
        return self._entities.get("house")

    @cached_property
    def activity_start_time(self) -> datetime:
        """Return the start time of the activity."""
        data = self._data
        return epoch_to_datetime(data.get("dateTime", data.get("timestamp")))

    @cached_property
    def activity_end_time(self) -> datetime:
        """Return the end time of the activity."""
        return self.activity_start_time

    @cached_property
    def action(self) -> str | None:
        """Return the action of the activity."""
        return self._data.get("action")

    @cached_property
    def device_id(self) -> str | None:
        """Return the ID of the device."""
        return self._data.get("deviceID")

    @cached_property
    def device_name(self) -> str | None:
        """Return the name of the device."""
        return self._data.get("deviceName")

    @cached_property
    def device_type(self) -> str | None:
        """Return the type of the device."""
        return self._data.get("deviceType")

    @cached_property
    def calling_user(self) -> dict[str, Any]:
        """Return the calling user."""
        return self._data.get("callingUser", self._data.get("user", {}))

    @cached_property
    def is_status(self) -> bool:
        """Return if the activity is a status update."""
        # If action is explicitly "status", it's a status update
        if self.action == "status":
            return True
        # If there's a manual operation (Bluetooth lock/unlock), it's NOT a status update
        user_id = self.calling_user.get("UserID", "")
        if user_id and user_id.startswith("manual"):
            return False
        # Check if this was marked as a manual operation (from PubNub without timestamp)
        if self._info.get("manual"):
            return False
        # Empty info typically means status update (except for WebSocket activities)
        return not self._info and self.source != SOURCE_WEBSOCKET


class BaseDoorbellMotionActivity(Activity):
    """Base class for doorbell motion activities."""

    def __repr__(self):
        return (
            f"<{self.__class__.__name__} action={self.action} activity_type={self.activity_type} "
            f"activity_start_time={self.activity_start_time} "
            f"device_name={self.device_name}"
            f"image_url={self.image_url}>"
            f"content_token={self.content_token}"
        )

    @cached_property
    def _image(self) -> dict[str, Any] | None:
        """Return the image of the activity."""
        return self._info.get("image")

    @cached_property
    def _content_token(self) -> str | None:
        """Return the content token of the activity."""
        return self._data.get("doorbell", {}).get("contentToken")

    @cached_property
    def image_url(self):
        """Return the image URL of the activity."""
        image = self._image
        return (None if image is None else image.get("secure_url")) or self._data.get(
            "attachment"
        )

    @cached_property
    def content_token(self):
        """Return the contentToken for the image URL"""
        return self._content_token or ""

    @cached_property
    def image_created_at_datetime(self):
        """Return the image created at datetime."""
        image = self._image
        if image is None:
            return None
        if "created_at" in image:
            return parse_datetime(image["created_at"])
        return self.activity_start_time


class DoorbellMotionActivity(BaseDoorbellMotionActivity):
    """A motion activity."""

    _activity_type = ActivityType.DOORBELL_MOTION


class DoorbellImageCaptureActivity(BaseDoorbellMotionActivity):
    """A motion activity with an image."""

    _activity_type = ActivityType.DOORBELL_IMAGE_CAPTURE


class DoorbellBaseActionActivity(Activity):
    """Base class for doorbell action activities."""

    @cached_property
    def image_url(self):
        """Return the image URL of the activity."""
        return self._info.get("image") or self._info.get("attachment")

    @cached_property
    def activity_start_time(self):
        """Return the start time of the activity."""
        if started := self._info.get("started"):
            return epoch_to_datetime(started)
        return super().activity_start_time

    @cached_property
    def activity_end_time(self):
        """Return the end time of the activity."""
        if ended := self._info.get("ended"):
            return epoch_to_datetime(ended)
        return super().activity_start_time


class DoorbellDingActivity(DoorbellBaseActionActivity):
    """Doorbell ding activity."""

    _activity_type = ActivityType.DOORBELL_DING


class DoorbellViewActivity(DoorbellBaseActionActivity):
    """Doorbell view activity."""

    _activity_type = ActivityType.DOORBELL_VIEW


class LockOperationActivity(Activity):
    """Lock operation activity."""

    _activity_type = ActivityType.LOCK_OPERATION_WITHOUT_OPERATOR

    def __init__(self, source: str, data: dict[str, Any]) -> None:
        """Initialize lock operation activity."""
        super().__init__(source, data)
        operated_by: str | None = None
        calling_user = self.calling_user
        first_name: str | None = calling_user.get("FirstName")
        last_name: str | None = calling_user.get("LastName")

        yale_user = self.yale_user
        if yale_user and first_name is None and last_name is None:
            first_name = yale_user.first_name
            last_name = yale_user.last_name

        # For legacy compatibility, we need to set the first_name and last_name
        # if its a physical or rf lock operation
        if (
            first_name is None
            and last_name is None
            and (first_last := ACTIVITY_TO_FIRST_LAST_NAME.get(self.action))
        ):
            first_name, last_name = first_last

        if first_name and last_name:
            operated_by = f"{first_name} {last_name}"
            self._activity_type = ActivityType.LOCK_OPERATION

        self._operated_by = operated_by

    @cached_property
    def _operator_image_urls(self) -> tuple[str | None, str | None]:
        """Return the image URLs of the lock operator."""
        operator_image_url: str | None = None
        operator_thumbnail_url: str | None = None
        calling_user = self.calling_user

        image_info = calling_user.get("imageInfo") or calling_user
        original = image_info.get("original")

        if type(original) is str:  # pylint: disable=unidiomatic-typecheck
            operator_image_url = original
        elif type(original) is dict:  # pylint: disable=unidiomatic-typecheck
            operator_image_url = original.get("secure_url")
        else:
            operator_image_url = None

        thumbnail = image_info.get("thumbnail")
        if type(thumbnail) is str:  # pylint: disable=unidiomatic-typecheck
            operator_thumbnail_url = thumbnail
        elif type(thumbnail) is dict:  # pylint: disable=unidiomatic-typecheck
            operator_thumbnail_url = thumbnail.get("secure_url")
        else:
            operator_thumbnail_url = None

        yale_user = self.yale_user
        if yale_user and not operator_image_url and not operator_thumbnail_url:
            operator_image_url = yale_user.image_url
            operator_thumbnail_url = yale_user.thumbnail_url

        if not operator_thumbnail_url:
            if icon := self._data.get("icon"):
                operator_thumbnail_url = icon

        if operator_thumbnail_url and not operator_image_url:
            operator_image_url = operator_thumbnail_url
        if operator_image_url and not operator_thumbnail_url:
            operator_thumbnail_url = operator_image_url

        return operator_image_url, operator_thumbnail_url

    def __repr__(self):
        """Return the representation."""
        return (
            f"<{self.__class__.__name__} action={self.action} activity_type={self.activity_type} "
            f"activity_start_time={self.activity_start_time} "
            f"device_name={self.device_name} "
            f"operated_by={self.operated_by} "
            f"operated_remote={self.operated_remote} "
            f"operated_keypad={self.operated_keypad} "
            f"operated_tag={self.operated_tag} "
            f"operated_manual={self.operated_manual} "
            f"operated_autorelock={self.operated_autorelock} "
            f"operator_image_url={self.operator_image_url} "
            f"operator_thumbnail_url={self.operator_thumbnail_url}>"
        )

    @cached_property
    def yale_user(self) -> YaleUser | None:
        """Return the Yale user."""
        return get_user_info(self.user_id)

    @cached_property
    def user_id(self) -> str | None:
        """Return the ID of the user."""
        return self.calling_user.get("UserID")

    @cached_property
    def operated_by(self):
        return self._operated_by

    @cached_property
    def operated_remote(self):
        """Operation was remote."""
        return self._info.get("remote", self.action in REMOTE_ACTIONS)

    @cached_property
    def operated_keypad(self):
        """Operation used keypad."""
        return self._info.get("keypad", self.action in KEYPAD_ACTIONS)

    @cached_property
    def operated_manual(self):
        """Operation done manually using the knob."""
        return self._info.get("manual", self.action in MANUAL_ACTIONS)

    @cached_property
    def operated_tag(self):
        """Operation used rfid tag."""
        return self._info.get("tag", self.action in TAG_ACTIONS)

    @cached_property
    def operated_autorelock(self):
        """Operation done by automatic relock."""
        return self.user_id == "automaticrelock" or self.action in AUTO_RELOCK_ACTIONS

    @cached_property
    def operator_image_url(self):
        """URL to the image of the lock operator."""
        return self._operator_image_urls[0]

    @cached_property
    def operator_thumbnail_url(self):
        """URL to the thumbnail of the lock operator."""
        return self._operator_image_urls[1]


class DoorOperationActivity(Activity):
    """Door operation activity."""

    _activity_type = ActivityType.DOOR_OPERATION


class BridgeOperationActivity(Activity):
    """Bridge operation activity."""

    _activity_type = ActivityType.BRIDGE_OPERATION


ActivityTypes = Union[
    DoorbellDingActivity,
    DoorbellMotionActivity,
    DoorbellImageCaptureActivity,
    DoorbellViewActivity,
    LockOperationActivity,
    DoorOperationActivity,
    BridgeOperationActivity,
]

ACTIONS_TO_CLASS = (
    (ACTIVITY_ACTIONS_DOORBELL_DING, DoorbellDingActivity),
    (ACTIVITY_ACTIONS_DOORBELL_MOTION, DoorbellMotionActivity),
    (ACTIVITY_ACTIONS_DOORBELL_IMAGE_CAPTURE, DoorbellImageCaptureActivity),
    (ACTIVITY_ACTIONS_DOORBELL_VIEW, DoorbellViewActivity),
    (ACTIVITY_ACTIONS_LOCK_OPERATION, LockOperationActivity),
    (ACTIVITY_ACTIONS_DOOR_OPERATION, DoorOperationActivity),
    (ACTIVITY_ACTIONS_BRIDGE_OPERATION, BridgeOperationActivity),
)

ACTION_TO_CLASS: dict[str, ActivityTypes] = {}
for activities, klass in ACTIONS_TO_CLASS:
    for activity in activities:
        ACTION_TO_CLASS[activity] = klass



from __future__ import annotations

import sys
from asyncio import AbstractEventLoop, Task, get_running_loop
from collections.abc import Coroutine
from typing import Any, TypeVar

_T = TypeVar("_T")

if sys.version_info >= (3, 12, 0):

    def create_eager_task(
        coro: Coroutine[Any, Any, _T],
        *,
        name: str | None = None,
        loop: AbstractEventLoop | None = None,
    ) -> Task[_T]:
        """Create a task from a coroutine and schedule it to run immediately."""
        return Task(coro, loop=loop or get_running_loop(), name=name, eager_start=True)

else:

    def create_eager_task(
        coro: Coroutine[Any, Any, _T],
        *,
        name: str | None = None,
        loop: AbstractEventLoop | None = None,
    ) -> Task[_T]:
        """Create a task from a coroutine and schedule it to run immediately."""
        return Task(
            coro,
            loop=loop or get_running_loop(),
            name=name,
        )



from unittest import mock

import pytest
from aiohttp import ClientResponseError

from yalexs.exceptions import (
    AugustApiAIOHTTPError,
    CannotConnect,
    InvalidAuth,
    RequireValidation,
    YaleApiError,
    YaleXSError,
)


def test_exceptions_can_be_empty_for_back_compat():
    assert InvalidAuth()
    assert str(InvalidAuth()) == "InvalidAuth"
    assert YaleApiError()
    assert str(YaleApiError()) == "YaleApiError"
    assert CannotConnect()
    assert str(CannotConnect()) == "CannotConnect"
    assert RequireValidation()
    assert YaleXSError()
    assert AugustApiAIOHTTPError()


def test_august_api_aio_http_error_reraise():
    mock_client_response_error = ClientResponseError(
        mock.MagicMock(),
        mock.MagicMock(),
        status=401,
    )
    ex = AugustApiAIOHTTPError("test", mock_client_response_error)
    assert str(ex) == "test"
    assert ex.auth_failed is True
    assert ex.aiohttp_client_error is mock_client_response_error
    assert ex.args == ("test",)


def test_subclassed_august_api_aio_http_error_reraise():
    mock_client_response_error = ClientResponseError(
        mock.MagicMock(),
        mock.MagicMock(),
        status=401,
    )
    with pytest.raises(InvalidAuth, match="test"):
        try:
            raise YaleApiError("test", mock_client_response_error)
        except AugustApiAIOHTTPError as ex:
            assert str(ex) == "test"
            assert ex.auth_failed is True
            assert ex.aiohttp_client_error is mock_client_response_error
            assert ex.args == ("test",)
            raise InvalidAuth(ex.args[0], ex) from ex



from ._compat import cached_property
from .device import DeviceDetail

BATTERY_LEVEL_FULL = "Full"
BATTERY_LEVEL_MEDIUM = "Medium"
BATTERY_LEVEL_LOW = "Low"

_LEVEL_TO_VALUE = {
    BATTERY_LEVEL_FULL: 100,
    BATTERY_LEVEL_MEDIUM: 60,
    BATTERY_LEVEL_LOW: 10,
}

_MAX_LEVEL = 200
_MIN_LEVEL = 120
_RATIO = 100 / (_MAX_LEVEL - _MIN_LEVEL)


class KeypadDetail(DeviceDetail):
    def __init__(self, house_id, keypad_name, data):
        super().__init__(
            data["_id"],
            keypad_name,
            house_id,
            data["serialNumber"],
            data["currentFirmwareVersion"],
            None,
            data,
        )
        self._battery_raw = data.get("batteryRaw")
        self._battery_level = data["batteryLevel"]

    @cached_property
    def model(self):
        return "AK-R1"

    @cached_property
    def battery_level(self):
        return self._battery_level

    @cached_property
    def battery_percentage(self):
        """Return an approximation of the battery percentage."""
        if self._battery_raw is not None:
            return int(max(0, min(100, (self._battery_raw - _MIN_LEVEL) * _RATIO)))
        if self._battery_level is None:
            return None
        return _LEVEL_TO_VALUE.get(self._battery_level, 0)



import asyncio
import logging
import sys
from collections.abc import Callable, Coroutine
from contextlib import suppress
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Any

import socketio

from ..api_common import api_auth_headers
from ..backports.tasks import create_eager_task
from ..const import Brand

if sys.version_info < (3, 11):
    UTC = timezone.utc
else:
    from datetime import UTC

_LOGGER = logging.getLogger(__name__)

if TYPE_CHECKING:
    from .gateway import Gateway

UpdateCallbackType = Callable[[str, datetime, dict[str, Any]], None]


class SocketIORunner:
    """Run the socketio client."""

    def __init__(self, gateway: "Gateway") -> None:
        """Initialize the socketio client."""
        self.gateway = gateway
        self._listeners: set[UpdateCallbackType] = set()
        self._access_token = None
        self.connected = False
        self._subscriber_id: str | None = None
        self._refresh_task: asyncio.Task | None = None

    def subscribe(self, callback: UpdateCallbackType) -> Callable[[], None]:
        """Add a listener."""
        self._listeners.add(callback)

        def _remove_listener():
            self._listeners.remove(callback)

        return _remove_listener

    def headers(self) -> dict[str, str]:
        """Get the headers."""
        return api_auth_headers(self._access_token, brand=Brand.YALE_GLOBAL)

    async def _refresh_access_token(self) -> None:
        """Refresh the access token."""
        self._access_token = await self.gateway.async_get_access_token()

    async def _run(self) -> None:
        """Run the socketio client."""
        sio = socketio.AsyncClient()

        @sio.event
        def connect() -> None:
            _LOGGER.debug("websocket connection established")
            self.connected = True

        @sio.event
        def data(data: dict[str, Any]) -> None:
            _LOGGER.debug("message received with %s", data)
            now = datetime.now(UTC)
            device_id = data.get("lockID")
            for listener in self._listeners:
                listener(device_id, now, data)

        @sio.event
        def disconnect() -> None:
            _LOGGER.debug("disconnected from server")
            self._refresh_task = create_eager_task(self._refresh_access_token())
            self.connected = False

        await sio.connect(
            f"https://websocket.aaecosystem.com/?subscriberID={self._subscriber_id}",
            retry=True,
            transports=["websocket"],
            headers=self.headers,
        )
        await sio.wait()

    async def run(
        self, user_uuid: str, brand: Brand = Brand.YALE_GLOBAL
    ) -> Callable[[], Coroutine[Any, Any, None]]:
        """Create a socketio session."""
        self._access_token = await self.gateway.async_get_access_token()
        api = self.gateway.api
        sub_info = await api.async_add_websocket_subscription(self._access_token)
        _LOGGER.debug("sub_info: %s", sub_info)
        self._subscriber_id = sub_info["subscriberID"]
        _LOGGER.debug("subscriberID: %s", self._subscriber_id)
        socketio_task = create_eager_task(self._run())

        async def _async_unsub():
            _LOGGER.debug("Shutting down socketio")
            socketio_task.cancel()
            self._listeners.clear()
            with suppress(asyncio.CancelledError):
                await socketio_task
            _LOGGER.debug("socketio stopped")

        return _async_unsub



"""Exceptions for errors."""

from __future__ import annotations

from ..exceptions import (  # noqa: F401
    CannotConnect,
    InvalidAuth,
    RequireValidation,
    YaleApiError,
    YaleXSError,
)



"""Tests for device capabilities functionality."""

from typing import Any

import pytest
from aiohttp import ClientSession
from aioresponses import aioresponses

from yalexs.api_async import ApiAsync
from yalexs.api_common import API_GET_CAPABILITIES_URL, ApiCommon
from yalexs.capabilities import CapabilitiesResponse
from yalexs.const import DEFAULT_BRAND
from yalexs.lock import LockDetail

ACCESS_TOKEN = "test-token"


@pytest.mark.asyncio
async def test_async_get_device_capabilities() -> None:
    """Test fetching device capabilities from the API."""
    capabilities_response: CapabilitiesResponse = {
        "lock": {
            "concurrentBLE": 2,
            "batteryType": "AA",
            "doorSense": True,
            "hasMagnetometer": False,
            "hasIntegratedWiFi": False,
            "scheduledSmartAlerts": True,
            "standalone": False,
            "bluetooth": True,
            "slotRange": None,
            "integratedKeypad": True,
            "entryCodeSlots": True,
            "pinSlotMax": 100,
            "pinSlotMin": 1,
            "supportsRFID": True,
            "supportsRFIDLegacy": False,
            "supportsRFIDCredential": True,
            "supportsRFIDOnlyAccess": True,
            "supportsRFIDWithCode": False,
            "supportsSecureMode": False,
            "supportsSecureModeCodeDisable": False,
            "supportsSecureModeMobileControl": False,
            "supportsFingerprintCredential": True,
            "supportsDeliveryMode": False,
            "supportsSchedulePerUser": True,
            "supportsFingerprintOnlyAccess": True,
            "batteryLifeMS": 21513600000,
            "supportedPartners": [],
            "unlatch": True,
        }
    }

    serial_number = "TEST123"

    with aioresponses() as mock:
        mock.get(
            ApiCommon(DEFAULT_BRAND).get_brand_url(API_GET_CAPABILITIES_URL)
            + f"?serialNumber={serial_number}&topLevelHost=true",
            payload=capabilities_response,
        )

        async with ClientSession() as session:
            api = ApiAsync(session)
            capabilities = await api.async_get_lock_capabilities(
                ACCESS_TOKEN, serial_number
            )

            assert capabilities == capabilities_response
            assert capabilities["lock"]["unlatch"] is True


def test_lock_detail_unlatch_supported_with_capabilities() -> None:
    """Test that LockDetail uses capabilities for unlatch_supported when available."""
    lock_data: dict[str, Any] = {
        "LockID": "test-lock-id",
        "LockName": "Test Lock",
        "HouseID": "test-house",
        "SerialNumber": "ABC123",
        "currentFirmwareVersion": "1.0.0",
        "Type": 5,  # Type that doesn't normally support unlatch
        "battery": 0.85,
        "LockStatus": {"status": "locked", "doorState": "closed"},
    }

    # Create lock detail without capabilities
    lock_detail = LockDetail(lock_data)

    # Should be False based on Type
    assert lock_detail.unlatch_supported is False

    # Set capabilities that indicate unlatch is supported
    capabilities: CapabilitiesResponse = {
        "lock": {
            "unlatch": True,
            "doorSense": True,
            "batteryType": "AA",
        }
    }
    lock_detail.set_capabilities(capabilities)

    # Now should be True based on capabilities
    assert lock_detail.unlatch_supported is True


def test_lock_detail_unlatch_supported_fallback_to_type() -> None:
    """Test that LockDetail falls back to Type-based check when no capabilities."""
    lock_data: dict[str, Any] = {
        "LockID": "test-lock-id",
        "LockName": "Test Lock",
        "HouseID": "test-house",
        "SerialNumber": "ABC123",
        "currentFirmwareVersion": "1.0.0",
        "Type": 17,  # Type 17 supports unlatch
        "battery": 0.85,
        "LockStatus": {"status": "locked", "doorState": "closed"},
    }

    # Create lock detail without capabilities
    lock_detail = LockDetail(lock_data)

    # Should be True based on Type 17
    assert lock_detail.unlatch_supported is True


def test_lock_detail_unlatch_supported_capabilities_override() -> None:
    """Test that capabilities override Type-based unlatch support."""
    lock_data: dict[str, Any] = {
        "LockID": "test-lock-id",
        "LockName": "Test Lock",
        "HouseID": "test-house",
        "SerialNumber": "ABC123",
        "currentFirmwareVersion": "1.0.0",
        "Type": 17,  # Type 17 normally supports unlatch
        "battery": 0.85,
        "LockStatus": {"status": "locked", "doorState": "closed"},
    }

    # Create lock detail
    lock_detail = LockDetail(lock_data)

    # Should be True based on Type 17
    assert lock_detail.unlatch_supported is True

    # Set capabilities that indicate unlatch is NOT supported
    capabilities: CapabilitiesResponse = {
        "lock": {
            "unlatch": False,  # Override: not supported
            "doorSense": True,
            "batteryType": "AA",
        }
    }
    lock_detail.set_capabilities(capabilities)

    # Now should be False based on capabilities override
    assert lock_detail.unlatch_supported is False


def test_lock_detail_set_capabilities() -> None:
    """Test setting capabilities on a lock detail."""
    lock_data: dict[str, Any] = {
        "LockID": "test-lock-id",
        "LockName": "Test Lock",
        "HouseID": "test-house",
        "SerialNumber": "ABC123",
        "currentFirmwareVersion": "1.0.0",
        "Type": 5,
        "battery": 0.85,
        "LockStatus": {"status": "locked", "doorState": "closed"},
    }

    lock_detail = LockDetail(lock_data)

    # Initially no capabilities
    assert lock_detail._capabilities is None

    # Set capabilities
    capabilities: CapabilitiesResponse = {
        "lock": {
            "unlatch": True,
            "doorSense": True,
            "batteryType": "AA",
            "pinSlotMax": 100,
            "pinSlotMin": 1,
        }
    }
    lock_detail.set_capabilities(capabilities)

    # Verify capabilities are stored
    assert lock_detail._capabilities == capabilities
    assert lock_detail._capabilities["lock"]["unlatch"] is True



"""Constants."""

from __future__ import annotations

from dataclasses import dataclass

from .backports.enum import StrEnum


class Brand(StrEnum):
    AUGUST = "august"
    YALE_ACCESS = "yale_access"
    YALE_HOME = "yale_home"
    YALE_GLOBAL = "yale_global"  # requires OAuth with Home Assistant
    YALE_AUGUST = "yale_august"  # requires OAuth with Home Assistant


DEFAULT_BRAND = Brand.AUGUST


@dataclass
class BrandConfig:
    """Brand configuration."""

    name: str
    branding: str
    access_token_header: str
    api_key_header: str
    branding_header: str
    api_key: str
    supports_doorbells: bool
    supports_alarms: bool
    require_oauth: bool
    base_url: str
    configuration_url: str
    pubnub_subscribe_token: str | None
    pubnub_publish_token: str | None


HEADER_VALUE_API_KEY_OLD = "7cab4bbd-2693-4fc1-b99b-dec0fb20f9d4"
HEADER_VALUE_API_KEY = "d9984f29-07a6-816e-e1c9-44ec9d1be431"

HEADER_AUGUST_ACCESS_TOKEN = "x-august-access-token"  # nosec
HEADER_AUGUST_API_KEY = "x-august-api-key"  # nosec
HEADER_AUGUST_BRANDING = "x-august-branding"

HEADER_ACCESS_TOKEN = "x-access-token"  # nosec
HEADER_API_KEY = "x-api-key"  # nosec
HEADER_BRANDING = "x-branding"

BRAND_CONFIG: dict[Brand, BrandConfig] = {
    Brand.AUGUST: BrandConfig(
        name="August",
        branding="august",
        access_token_header=HEADER_AUGUST_ACCESS_TOKEN,
        api_key_header=HEADER_AUGUST_API_KEY,
        branding_header=HEADER_AUGUST_BRANDING,
        api_key=HEADER_VALUE_API_KEY,
        supports_doorbells=True,
        supports_alarms=False,
        require_oauth=False,
        base_url="https://api-production.august.com",
        configuration_url="https://account.august.com",
        pubnub_subscribe_token="sub-c-1030e062-0ebe-11e5-a5c2-0619f8945a4f",  # nosec
        pubnub_publish_token="pub-c-567d7f2d-270a-438a-a785-f0af12ad8312",  # nosec
    ),
    Brand.YALE_ACCESS: BrandConfig(
        name="Yale Access",
        branding="yale",
        access_token_header=HEADER_AUGUST_ACCESS_TOKEN,
        api_key_header=HEADER_AUGUST_API_KEY,
        branding_header=HEADER_AUGUST_BRANDING,
        api_key=HEADER_VALUE_API_KEY,
        supports_doorbells=True,
        supports_alarms=False,
        require_oauth=False,
        base_url="https://api-production.august.com",
        configuration_url="https://account.august.com",
        pubnub_subscribe_token="sub-c-1030e062-0ebe-11e5-a5c2-0619f8945a4f",  # nosec
        pubnub_publish_token="pub-c-567d7f2d-270a-438a-a785-f0af12ad8312",  # nosec
    ),
    Brand.YALE_HOME: BrandConfig(
        name="Yale Home",
        branding="yale",
        access_token_header=HEADER_ACCESS_TOKEN,
        api_key_header=HEADER_API_KEY,
        branding_header=HEADER_BRANDING,
        api_key="6e2a2093-6118-42c5-8a41-e1fd25dce7a1",  # 🤞
        supports_doorbells=True,
        supports_alarms=True,
        require_oauth=False,
        base_url="https://api.aaecosystem.com",
        configuration_url="https://account.aaecosystem.com",
        pubnub_subscribe_token="sub-c-c9c38d4d-5796-46c9-9262-af20cf6a1d42",  # nosec
        pubnub_publish_token="pub-c-353e8881-cf58-4b26-9baf-96f296de0677",  # nosec
    ),
    Brand.YALE_GLOBAL: BrandConfig(
        name="Yale Global",
        branding="yale",
        access_token_header=HEADER_ACCESS_TOKEN,
        api_key_header=HEADER_API_KEY,
        branding_header=HEADER_BRANDING,
        # Sadly we currently do not have a way to avoid
        # having the API key in the code because it must
        # run on the user's device
        api_key="d16a1029-d823-4b55-a4ce-a769a9b56f0e",
        supports_doorbells=True,
        supports_alarms=True,  # ??
        require_oauth=True,
        base_url="https://api.aaecosystem.com",
        configuration_url="https://account.aaecosystem.com",
        # This brand uses WebSockets and has migrated
        # away from PubNub for this purpose which is great
        # because its one less credential we have to expose
        # to the user
        pubnub_publish_token=None,
        pubnub_subscribe_token=None,
    ),
    Brand.YALE_AUGUST: BrandConfig(
        name="Yale August",
        branding="august",
        access_token_header=HEADER_AUGUST_ACCESS_TOKEN,
        api_key_header=HEADER_AUGUST_API_KEY,
        branding_header=HEADER_AUGUST_BRANDING,
        api_key="66814fd9-af2c-426c-9710-b37e7eadfb51",
        supports_doorbells=True,
        supports_alarms=False,
        require_oauth=True,
        base_url="https://api-production.august.com",
        configuration_url="https://account.august.com",
        pubnub_subscribe_token="sub-c-1030e062-0ebe-11e5-a5c2-0619f8945a4f",  # nosec
        pubnub_publish_token="pub-c-567d7f2d-270a-438a-a785-f0af12ad8312",  # nosec
    ),
}

BRANDS = {brand: brand_config.name for brand, brand_config in BRAND_CONFIG.items()}
BRANDS_WITHOUT_OAUTH = {
    brand: brand_config.name
    for brand, brand_config in BRAND_CONFIG.items()
    if not brand_config.require_oauth
}
BRANDING = {
    brand: brand_config.branding for brand, brand_config in BRAND_CONFIG.items()
}
BASE_URLS = {
    brand: brand_config.base_url for brand, brand_config in BRAND_CONFIG.items()
}
CONFIGURATION_URLS = {
    brand: brand_config.configuration_url
    for brand, brand_config in BRAND_CONFIG.items()
}



from __future__ import annotations

import datetime
import random
import ssl
from functools import cache
from typing import TYPE_CHECKING, Union

from .activity import (
    ACTION_BRIDGE_OFFLINE,
    ACTION_BRIDGE_ONLINE,
    ACTIVITY_ACTION_STATES,
    ACTIVITY_MOVING_STATES,
    MOVING_STATES,
    BridgeOperationActivity,
    DoorbellImageCaptureActivity,
    DoorbellMotionActivity,
    DoorOperationActivity,
    LockOperationActivity,
)
from .const import CONFIGURATION_URLS, Brand
from .lock import LockDetail

LockActivityTypes = Union[
    LockOperationActivity, DoorOperationActivity, BridgeOperationActivity
]
DoorbellActivityTypes = Union[
    DoorbellImageCaptureActivity, DoorbellMotionActivity, BridgeOperationActivity
]

if TYPE_CHECKING:
    from .doorbell import DoorbellDetail


def get_latest_activity(
    activity1: LockActivityTypes | None, activity2: LockActivityTypes | None
) -> LockActivityTypes | None:
    """Return the latest activity."""
    return (
        activity2
        if (
            not activity1
            or (
                activity2
                and activity2.action not in ACTIVITY_MOVING_STATES
                and activity1.activity_start_time <= activity2.activity_start_time
            )
        )
        else activity1
    )


def update_lock_detail_from_activity(
    lock_detail: LockDetail,
    activity: LockActivityTypes,
) -> bool:
    """Update the LockDetail from an activity."""
    activity_end_time_utc = as_utc_from_local(activity.activity_end_time)
    if activity.device_id != lock_detail.device_id:
        raise ValueError
    if isinstance(activity, LockOperationActivity):
        if lock_detail.lock_status_datetime and (
            lock_detail.lock_status_datetime > activity_end_time_utc
            or (
                lock_detail.lock_status_datetime == activity_end_time_utc
                and lock_detail.lock_status not in MOVING_STATES
            )
        ):
            return False
        lock_detail.lock_status = ACTIVITY_ACTION_STATES[activity.action]
        lock_detail.lock_status_datetime = activity_end_time_utc
    elif isinstance(activity, DoorOperationActivity):
        if (
            lock_detail.door_state_datetime
            and lock_detail.door_state_datetime >= activity_end_time_utc
        ):
            return False
        lock_detail.door_state = ACTIVITY_ACTION_STATES[activity.action]
        lock_detail.door_state_datetime = activity_end_time_utc
    elif isinstance(activity, BridgeOperationActivity):
        if activity.action == ACTION_BRIDGE_ONLINE:
            lock_detail.set_online(True)
        elif activity.action == ACTION_BRIDGE_OFFLINE:
            lock_detail.set_online(False)
    else:
        raise ValueError

    return True


def update_doorbell_image_from_activity(
    doorbell_detail: DoorbellDetail, activity: DoorbellActivityTypes
) -> bool:
    """Update the DoorDetail from an activity with a new image."""
    if activity.device_id != doorbell_detail.device_id:
        raise ValueError
    if isinstance(activity, (DoorbellImageCaptureActivity, DoorbellMotionActivity)):
        if activity.image_created_at_datetime is None:
            return False

        if (
            doorbell_detail.image_created_at_datetime is None
            or doorbell_detail.image_created_at_datetime
            < activity.image_created_at_datetime
        ):
            doorbell_detail.image_url = activity.image_url
            doorbell_detail.image_created_at_datetime = (
                activity.image_created_at_datetime
            )
            doorbell_detail.content_token = (
                activity.content_token or doorbell_detail.content_token
            )
        else:
            return False
    else:
        raise ValueError

    return True


def as_utc_from_local(dtime: datetime.datetime) -> datetime.datetime:
    """Converts the datetime returned from an activity to UTC."""
    return dtime.astimezone(tz=datetime.timezone.utc)


def get_configuration_url(brand: Brand) -> str:
    """Return the configuration URL for the brand."""
    return CONFIGURATION_URLS[brand]


@cache
def get_ssl_context() -> ssl.SSLContext:
    """Return an SSL context for cloudflare."""
    context = ssl.create_default_context()
    ciphers = [cipher["name"] for cipher in context.get_ciphers()]
    default_ciphers = ciphers[:3]
    backup_ciphers = ciphers[3:]
    random.shuffle(backup_ciphers)
    context.set_ciphers(":".join((*default_ciphers, *backup_ciphers)))
    return context



import logging
from datetime import datetime
from typing import Any

from .activity import (
    ACTION_BRIDGE_OFFLINE,
    ACTION_BRIDGE_ONLINE,
    ACTION_DOOR_CLOSED,
    ACTION_DOOR_OPEN,
    ACTION_DOORBELL_BUTTON_PUSHED,
    ACTION_DOORBELL_IMAGE_CAPTURE,
    ACTION_DOORBELL_MOTION_DETECTED,
    ACTION_LOCK_JAMMED,
    ACTION_LOCK_LOCK,
    ACTION_LOCK_LOCKING,
    ACTION_LOCK_UNLATCH,
    ACTION_LOCK_UNLATCHING,
    ACTION_LOCK_UNLOCK,
    ACTION_LOCK_UNLOCKING,
    SOURCE_PUBNUB,
    ActivityTypes,
)
from .api_common import _activity_from_dict, _datetime_string_to_epoch
from .device import Device
from .doorbell import DOORBELL_STATUS_KEY, DoorbellDetail
from .lock import (
    DOOR_STATE_KEY,
    LOCK_STATUS_KEY,
    LockDetail,
    LockDoorStatus,
    LockStatus,
    determine_door_state,
    determine_lock_status,
)

LOCK_STATUS_TO_ACTION = {
    LockStatus.LOCKED: ACTION_LOCK_LOCK,
    LockStatus.UNLATCHED: ACTION_LOCK_UNLATCH,
    LockStatus.UNLOCKED: ACTION_LOCK_UNLOCK,
    LockStatus.LOCKING: ACTION_LOCK_LOCKING,
    LockStatus.UNLATCHING: ACTION_LOCK_UNLATCHING,
    LockStatus.UNLOCKING: ACTION_LOCK_UNLOCKING,
    LockStatus.JAMMED: ACTION_LOCK_JAMMED,
}

_BRIDGE_ACTIONS = {ACTION_BRIDGE_ONLINE, ACTION_BRIDGE_OFFLINE}


_LOGGER = logging.getLogger(__name__)


def activities_from_pubnub_message(  # noqa: C901
    device: Device,
    date_time: datetime,
    message: dict[str, Any],
    source: str = SOURCE_PUBNUB,
) -> list[ActivityTypes]:
    """Create activities from pubnub."""
    activities: list[ActivityTypes] = []
    activity_dict = {
        "deviceID": device.device_id,
        "house": device.house_id,
        "deviceName": device.device_name,
    }
    info = message.get("info", {})
    context = info.get("context", {})
    accept_user = False
    if "startTime" in info:
        activity_dict["dateTime"] = _datetime_string_to_epoch(info["startTime"])
        accept_user = True
    elif "startDate" in context:
        activity_dict["dateTime"] = _datetime_string_to_epoch(context["startDate"])
        accept_user = True
    else:
        activity_dict["dateTime"] = date_time.timestamp() * 1000

    if isinstance(device, LockDetail):
        activity_dict["deviceType"] = "lock"
        activity_dict["info"] = info
        calling_user_id = message.get("callingUserID")

        # Some locks sometimes send lots of status messages, triggered by the app. Ignore these.
        if (
            info.get("action") == "status"
            and not message.get("error")
            and message.get("result") != "failed"
        ):
            _LOGGER.debug("Not creating lock activity from status push")
            return activities

        # Only accept a UserID if we have a date/time
        # as otherwise it is a duplicate of the previous
        # activity
        if accept_user and calling_user_id:
            activity_dict["callingUser"] = {"UserID": calling_user_id}
        # Mark manual operations so they're not treated as status updates
        elif calling_user_id and calling_user_id.startswith("manual"):
            activity_dict["info"]["manual"] = True
        if "remoteEvent" in message:
            activity_dict["info"]["remote"] = True
        error = message.get("error") or {}
        if error.get("restCode") == 98 or error.get("name") == "ERRNO_BRIDGE_OFFLINE":
            _add_activity(activities, activity_dict, ACTION_BRIDGE_OFFLINE, source)
        elif status := message.get("lockAction", message.get(LOCK_STATUS_KEY)):
            if status in _BRIDGE_ACTIONS:
                _add_activity(activities, activity_dict, status, source)
            if action := LOCK_STATUS_TO_ACTION.get(determine_lock_status(status)):
                _add_activity(activities, activity_dict, action, source)
        if door_state_raw := message.get(DOOR_STATE_KEY):
            door_state = determine_door_state(door_state_raw)
            if door_state == LockDoorStatus.OPEN:
                _add_activity(activities, activity_dict, ACTION_DOOR_OPEN, source)
            elif door_state == LockDoorStatus.CLOSED:
                _add_activity(activities, activity_dict, ACTION_DOOR_CLOSED, source)

    elif isinstance(device, DoorbellDetail):
        activity_dict["deviceType"] = "doorbell"
        info = activity_dict["info"] = message.get("data", {})
        info.setdefault("image", info.get("result", {}))
        info.setdefault("started", activity_dict["dateTime"])
        info.setdefault("ended", activity_dict["dateTime"])

        if (status := message.get(DOORBELL_STATUS_KEY)) and status in (
            ACTION_DOORBELL_MOTION_DETECTED,
            ACTION_DOORBELL_IMAGE_CAPTURE,
            ACTION_DOORBELL_BUTTON_PUSHED,
        ):
            _add_activity(activities, activity_dict, status, source)

    return activities


def _add_activity(
    activities: list[ActivityTypes],
    activity_dict: dict[str, Any],
    action: str,
    source: str,
) -> None:
    """Add an activity."""
    activity_dict = activity_dict.copy()
    activity_dict["action"] = action
    activities.append(
        _activity_from_dict(source, activity_dict, _LOGGER.isEnabledFor(logging.DEBUG))
    )



from __future__ import annotations

from enum import Enum
from typing import Any

from ._compat import cached_property
from .device import DeviceDetail


class BridgeStatus(Enum):
    OFFLINE = "offline"
    ONLINE = "online"
    UNKNOWN = "unknown"


class BridgeDetail(DeviceDetail):
    """Represents a bridge device."""

    def __init__(self, house_id: str, data: dict[str, Any]) -> None:
        """Initialize the bridge device."""
        super().__init__(
            data["_id"], None, house_id, None, data["firmwareVersion"], None, data
        )

        self._hyper_bridge = data.get("hyperBridge", False)
        self._operative = data["operative"]

        if "status" in data:
            self._status = BridgeStatusDetail(data["status"])
        else:
            self._status = None

    @property
    def status(self):
        return self._status

    @cached_property
    def hyper_bridge(self):
        return self._hyper_bridge

    @cached_property
    def operative(self):
        return self._operative

    def set_online(self, state):
        """Called when the bridge online state changes."""
        self._status.set_online(state)


class BridgeStatusDetail:
    """Represents the status of a bridge device."""

    def __init__(self, data: dict[str, Any]) -> None:
        """Initialize the bridge status."""
        self._current = BridgeStatus.UNKNOWN

        if "current" in data and data["current"] == "online":
            self._current = BridgeStatus.ONLINE

        self._updated = data.get("updated")
        self._last_online = data.get("lastOnline")
        self._last_offline = data.get("lastOffline")

    @property
    def current(self):
        return self._current

    def set_online(self, state):
        """Called when the bridge online state changes."""
        self._current = BridgeStatus.ONLINE if state else BridgeStatus.OFFLINE

    @cached_property
    def updated(self):
        return self._updated

    @cached_property
    def last_online(self):
        return self._last_online

    @cached_property
    def last_offline(self):
        return self._last_offline



from __future__ import annotations

import datetime
from functools import lru_cache

import ciso8601
import dateutil.parser


@lru_cache(maxsize=512)
def epoch_to_datetime(epoch: str | float) -> datetime.datetime:
    """Convert epoch to datetime."""
    return datetime.datetime.fromtimestamp(float(epoch) / 1000.0)


def parse_datetime(datetime_string: str) -> datetime.datetime:
    """Parse a datetime string."""
    try:
        return ciso8601.parse_datetime(datetime_string)
    except ValueError:
        return dateutil.parser.parse(datetime_string)



from __future__ import annotations

import json
import logging
from datetime import datetime, timedelta, timezone

import aiofiles
from aiohttp import ClientError

from .api_async import ApiAsync
from .authenticator_common import (
    Authentication,
    AuthenticationState,
    AuthenticatorCommon,
    ValidationResult,
    from_authentication_json,
    to_authentication_json,
)
from .exceptions import AugustApiAIOHTTPError

_LOGGER = logging.getLogger(__name__)


class AuthenticatorAsync(AuthenticatorCommon):
    """Class to manage authentication with the August API."""

    _api: ApiAsync

    async def _read_access_token_file(
        self,
        access_token_cache_file: str,
        file: aiofiles.threadpool.binary.AsyncBufferedIOBase,
    ) -> None:
        contents = await file.read()
        self._authentication = from_authentication_json(json.loads(contents))

        # If token is to expire within 7 days then print a warning.
        if self._authentication.is_expired():
            _LOGGER.error("Token has expired.")
            self._authentication = Authentication(
                AuthenticationState.REQUIRES_AUTHENTICATION,
                install_id=self._install_id,
            )
        # If token is not expired but less then 7 days before it
        # will.
        elif (
            self._authentication.parsed_expiration_time() - datetime.now(timezone.utc)
        ) < timedelta(days=7):
            exp_time = self._authentication.access_token_expires
            _LOGGER.warning(
                "API Token is going to expire at %s "
                "hours. Deleting file %s will result "
                "in a new token being requested next"
                " time",
                exp_time,
                access_token_cache_file,
            )

    async def async_setup_authentication(self) -> None:
        if access_token_cache_file := self._access_token_cache_file:
            try:
                async with aiofiles.open(access_token_cache_file) as file:
                    await self._read_access_token_file(access_token_cache_file, file)
                    return
            except FileNotFoundError:
                _LOGGER.debug("Cache file not found: %s", access_token_cache_file)
            except json.decoder.JSONDecodeError as error:
                _LOGGER.error(
                    "Unable to read cache file (%s): %s",
                    access_token_cache_file,
                    error,
                )

        self._authentication = Authentication(
            AuthenticationState.REQUIRES_AUTHENTICATION, install_id=self._install_id
        )

    async def async_authenticate(self) -> Authentication:
        if self._api.brand_config.require_oauth:
            raise RuntimeError(f"OAuth is required for brand {self._api.brand}")

        if self._authentication.state is AuthenticationState.AUTHENTICATED:
            return self._authentication

        identifier = self._login_method + ":" + self._username
        install_id = self._authentication.install_id
        response = await self._api.async_get_session(
            install_id, identifier, self._password
        )

        json_dict = await response.json()
        authentication = self._authentication_from_session_response(
            install_id, response.headers, json_dict
        )

        if authentication.state is AuthenticationState.AUTHENTICATED:
            await self._async_cache_authentication(authentication)

        return authentication

    async def async_validate_verification_code(
        self, verification_code: str
    ) -> ValidationResult:
        if not verification_code:
            return ValidationResult.INVALID_VERIFICATION_CODE

        try:
            await self._api.async_validate_verification_code(
                self._authentication.access_token,
                self._login_method,
                self._username,
                verification_code,
            )
        except (AugustApiAIOHTTPError, ClientError):
            return ValidationResult.INVALID_VERIFICATION_CODE

        return ValidationResult.VALIDATED

    async def async_send_verification_code(self) -> bool:
        await self._api.async_send_verification_code(
            self._authentication.access_token, self._login_method, self._username
        )

        return True

    async def async_refresh_access_token(self, force=False) -> Authentication | None:
        if not self.should_refresh() and not force:
            return self._authentication

        if self._authentication.state is not AuthenticationState.AUTHENTICATED:
            _LOGGER.warning("Tried to refresh access token when not authenticated")
            return self._authentication

        refreshed_token = await self._api.async_refresh_access_token(
            self._authentication.access_token
        )

        authentication = self._process_refreshed_access_token(refreshed_token)
        await self._async_cache_authentication(authentication)
        return authentication

    async def _async_cache_authentication(self, authentication: Authentication) -> None:
        if self._access_token_cache_file is not None:
            async with aiofiles.open(self._access_token_cache_file, "w") as file:
                await file.write(to_authentication_json(authentication))



from __future__ import annotations

import asyncio
import time
from asyncio import AbstractEventLoop, TimerHandle
from typing import Union

_MONOTONIC_RESOLUTION = time.get_clock_info("monotonic").resolution
ScheduledType = Union[TimerHandle, tuple[float, TimerHandle]]


def get_scheduled_timer_handles(loop: AbstractEventLoop) -> list[TimerHandle]:
    """Return a list of scheduled TimerHandles."""
    handles: list[ScheduledType] = loop._scheduled  # type: ignore[attr-defined]
    return [
        handle if isinstance(handle, TimerHandle) else handle[1] for handle in handles
    ]


def fire_time_changed() -> None:
    timestamp = time.time()
    loop = asyncio.get_running_loop()
    for task in list(get_scheduled_timer_handles(loop)):
        if not isinstance(task, asyncio.TimerHandle):
            continue
        if task.cancelled():
            continue

        mock_seconds_into_future = timestamp - time.time()
        future_seconds = task.when() - (loop.time() + _MONOTONIC_RESOLUTION)

        if mock_seconds_into_future >= future_seconds:
            task._run()
            task.cancel()



"""Consume the august activity stream."""

from __future__ import annotations

import asyncio
import logging
from collections import defaultdict

from aiohttp import ClientError

from ..activity import Activity, ActivityType
from ..api_async import ApiAsync
from ..backports.tasks import create_eager_task
from ..exceptions import AugustApiAIOHTTPError
from ..pubnub_async import AugustPubNub
from ..util import get_latest_activity
from .const import ACTIVITY_UPDATE_INTERVAL
from .gateway import Gateway
from .socketio import SocketIORunner
from .subscriber import SubscriberMixin

_LOGGER = logging.getLogger(__name__)

ACTIVITY_STREAM_FETCH_LIMIT = 10
ACTIVITY_CATCH_UP_FETCH_LIMIT = 2500

INITIAL_LOCK_RESYNC_TIME = 45

# If there is a storm of activity (ie lock, unlock, door open, door close, etc)
# we want to debounce the updates so we don't hammer the activity api too much.
ACTIVITY_DEBOUNCE_COOLDOWN = 4.0

# How long we expect it to take between when we get a WebSocket/PubNub
# message and the activity API to be updated.
UPDATE_SOON = 3.0

NEVER_TIME = -86400.0


class ActivityStream(SubscriberMixin):
    """August activity stream handler."""

    def __init__(
        self,
        api: ApiAsync,
        august_gateway: Gateway,
        house_ids: set[str],
        push: AugustPubNub | SocketIORunner,
    ) -> None:
        """Init activity stream object."""
        super().__init__(ACTIVITY_UPDATE_INTERVAL)
        self._schedule_updates: dict[str, asyncio.TimerHandle] = {}
        self._august_gateway = august_gateway
        self._api = api
        self._house_ids = house_ids
        self._latest_activities: defaultdict[
            str, dict[ActivityType, Activity | None]
        ] = defaultdict(lambda: defaultdict(lambda: None))
        self._did_first_update = False
        self.push = push
        self._update_tasks: dict[str, asyncio.Task] = {}
        self._last_update_time: dict[str, float] = dict.fromkeys(house_ids, NEVER_TIME)
        self._start_time: float | None = None
        self._pending_updates: dict[str, int] = dict.fromkeys(house_ids, 1)
        self._loop = asyncio.get_running_loop()
        self._shutdown: bool = False

    async def async_setup(self) -> None:
        """Token refresh check and catch up the activity stream."""
        self._start_time = self._loop.time()
        await self._async_refresh()
        await self._async_first_refresh()
        self._did_first_update = True

    def async_stop(self) -> None:
        """Cleanup any debounces."""
        self._shutdown = True
        for task in self._update_tasks.values():
            task.cancel()
        self._update_tasks.clear()
        self._async_cancel_all_future_updates()

    def _async_cancel_future_updates(self, house_id: str) -> None:
        """Cancel future updates."""
        if handle := self._schedule_updates.pop(house_id, None):
            handle.cancel()
        self._pending_updates[house_id] = 0

    def _async_cancel_all_future_updates(self) -> None:
        """Cancel all future updates."""
        for house_id in self._house_ids:
            self._async_cancel_future_updates(house_id)

    def get_latest_device_activity(
        self, device_id: str, activity_types: set[ActivityType]
    ) -> Activity | None:
        """Return latest activity that is one of the activity_types."""
        if not (latest_device_activities := self._latest_activities.get(device_id)):
            return None

        latest_activity: Activity | None = None

        for activity_type in activity_types:
            if activity := latest_device_activities.get(activity_type):
                if (
                    latest_activity
                    and activity.activity_start_time
                    <= latest_activity.activity_start_time
                ):
                    continue
                latest_activity = activity

        return latest_activity

    @property
    def push_updates_connected(self) -> bool:
        """Return if the push updates are connected."""
        return self.push.connected

    async def _async_refresh(self) -> None:
        """Update the activity stream from August."""
        # This is the only place we refresh the api token
        if self._shutdown:
            return
        await self._august_gateway.async_refresh_access_token_if_needed()
        if not self.push_updates_connected:
            _LOGGER.debug("Push updates are not connected, data will be stale")

    async def _async_first_refresh(self) -> None:
        """Update the activity stream from August for the first time."""
        if self.push_updates_connected:
            _LOGGER.debug("Skipping update because push updates are active")
            return
        _LOGGER.debug("Start retrieving device activities")
        # Await in sequence to avoid hammering the API
        for house_id in self._house_ids:
            if not self._update_running(house_id):
                await self._create_update_task(house_id)

    def _create_update_task(self, house_id: str) -> asyncio.Task:
        """Create an update task."""
        if self._update_running(house_id):
            raise RuntimeError("Update already running")
        self._update_tasks[house_id] = create_eager_task(
            self._async_execute_schedule_update(house_id), loop=self._loop
        )
        return self._update_tasks[house_id]

    def _update_running(self, house_id: str) -> bool:
        """Return if an update is running for the house id."""
        return bool(
            (current_task := self._update_tasks.get(house_id))
            and not current_task.done()
        )

    def _updated_recently(self, house_id: str, now: float) -> bool:
        """Return if the house id was updated recently."""
        return self._last_update_time[house_id] + ACTIVITY_DEBOUNCE_COOLDOWN > now

    def _async_schedule_update_callback(self, house_id: str) -> None:
        """Schedule an update callback."""
        self._schedule_updates.pop(house_id, None)
        now = self._loop.time()
        if delay := self._determine_update_delay(house_id, now, from_callback=True):
            self._async_schedule_update(house_id, now, delay)
            return
        self._create_update_task(house_id)

    def _determine_update_delay(
        self, house_id: str, now: float, from_callback: bool = False
    ) -> float:
        """Return if we should delay the update."""
        if self._updated_recently(house_id, now) or self._update_running(house_id):
            return ACTIVITY_DEBOUNCE_COOLDOWN
        if not self._initial_resync_complete(now):
            return INITIAL_LOCK_RESYNC_TIME
        return 0 if from_callback else UPDATE_SOON

    def _async_schedule_update(self, house_id: str, now: float, delay: float) -> None:
        """Update the activity stream now or in the future if its too soon."""
        if self._shutdown or self._pending_updates[house_id] <= 0:
            return
        _LOGGER.debug(
            "Scheduling update for house id %s in %s seconds", house_id, delay
        )
        # Do not update right away because the activities API is
        # likely not updated yet and we will just get the same
        # activities again. Instead, schedule the update for
        # the future.
        if scheduled := self._schedule_updates.pop(house_id, None):
            scheduled.cancel()
        self._schedule_updates[house_id] = self._loop.call_at(
            now + delay, self._async_schedule_update_callback, house_id
        )

    async def _async_execute_schedule_update(self, house_id: str) -> None:
        """Execute a scheduled update."""
        self._pending_updates[house_id] -= 1
        self._last_update_time[house_id] = self._loop.time()
        await self._async_update_house_id(house_id)
        if (pending_count := self._pending_updates[house_id]) > 0:
            _LOGGER.debug(
                "There are %s pending updates for house id %s", pending_count, house_id
            )
            now = self._loop.time()
            delay = self._determine_update_delay(house_id, now)
            self._async_schedule_update(house_id, now, delay)

    def _initial_resync_complete(self, now: float) -> bool:
        """Return if the initial resync is complete."""
        return self._start_time and now - self._start_time > INITIAL_LOCK_RESYNC_TIME

    def _set_update_count(self, house_id: str, now: float) -> None:
        """Set the update count."""
        # Schedule one update soon and two updates past the debounce time
        # to ensure we catch the case where the activity
        # api does not update right away and we need to poll
        # it again. Sometimes the lock operator or a doorbell
        # will not show up in the activity stream right away.
        # Only do additional polls if we are past
        # the initial lock resync time to avoid a storm
        # of activity at setup.
        if not self._initial_resync_complete(now):
            # No resync yet, above spamming the API
            update_count = 1
        else:
            # Initial resync complete, we can do 2 updates
            update_count = 2
        self._pending_updates[house_id] = update_count

    def async_schedule_house_id_refresh(self, house_id: str) -> None:
        """Update for a house activities now and once in the future."""
        self._async_cancel_future_updates(house_id)
        now = self._loop.time()
        self._set_update_count(house_id, now)
        delay = self._determine_update_delay(house_id, now)
        self._async_schedule_update(house_id, now, delay)

    def _activity_limit(self) -> bool:
        """Return if the activity limit has been reached."""
        if self._did_first_update:
            return ACTIVITY_STREAM_FETCH_LIMIT
        return ACTIVITY_CATCH_UP_FETCH_LIMIT

    async def _async_update_house_id(self, house_id: str) -> None:
        """Update device activities for a house.

        Must only be called from _async_execute_schedule_update
        """
        if self._shutdown:
            return

        _LOGGER.debug("Updating device activity for house id %s", house_id)
        try:
            activities = await self._api.async_get_house_activities(
                await self._august_gateway.async_get_access_token(),
                house_id,
                limit=self._activity_limit(),
            )
        except (AugustApiAIOHTTPError, ClientError) as ex:
            _LOGGER.error(
                "Request error trying to retrieve activity for house id %s: %s",
                house_id,
                ex,
            )
            # Make sure we process the next house if one of them fails
            return

        _LOGGER.debug(
            "Completed retrieving device activities for house id %s", house_id
        )
        for device_id in self.async_process_newer_device_activities(activities):
            _LOGGER.debug(
                "async_signal_device_id_update (from activity stream): %s",
                device_id,
            )
            self.async_signal_device_id_update(device_id)

    def async_process_newer_device_activities(
        self, activities: list[Activity]
    ) -> set[str]:
        """Process activities if they are newer than the last one."""
        updated_device_ids: set[str] = set()
        latest_activities = self._latest_activities
        for activity in activities:
            device_id = activity.device_id
            activity_type = activity.activity_type
            device_activities = latest_activities[device_id]
            # Ignore activities that are older than the latest one unless it is a non
            # locking or unlocking activity with the exact same start time.
            last_activity = device_activities[activity_type]
            # The activity stream can have duplicate activities. So we need
            # to call get_latest_activity to figure out if if the activity
            # is actually newer than the last one.
            latest_activity = get_latest_activity(activity, last_activity)
            if latest_activity != activity:
                _LOGGER.debug(
                    "Skipping activity %s for device %s as it is not newer than the last one: %s",
                    activity,
                    device_id,
                    last_activity,
                )
                continue

            device_activities[activity_type] = activity
            updated_device_ids.add(device_id)

        return updated_device_ids



"""Test the manager data module."""

import logging
from datetime import datetime
from typing import Any
from unittest.mock import AsyncMock, Mock, patch

import pytest
from aiohttp import ClientError, ClientResponseError

from yalexs.activity import SOURCE_PUBNUB, SOURCE_WEBSOCKET
from yalexs.capabilities import CapabilitiesResponse
from yalexs.const import Brand
from yalexs.exceptions import YaleApiError
from yalexs.lock import LockDetail, LockOperation
from yalexs.manager.data import YaleXSData


class MockYaleXSData(YaleXSData):
    """Mock implementation of YaleXSData with mocked abstract method."""

    def async_offline_key_discovered(self, detail) -> None:
        """Mock implementation of abstract method."""


class TestPushStateTracking:
    """Test push state tracking functionality."""

    def setup_method(self):
        """Set up test fixtures."""

        # Create a simple object with just the method and state dict we need
        class TestData:
            def __init__(self):
                self._last_push_state = {}

            # Bind the actual method from YaleXSData
            _is_unchanged_push_state = YaleXSData._is_unchanged_push_state

        self.data = TestData()
        self.device_id = "test_device_id"

    def test_pubnub_initial_status_update_sets_baseline(self):
        """Test that the first PubNub status update sets the baseline state."""
        # First status update should set baseline
        message1 = {
            "status": "locked",
            "doorState": "closed",
        }

        # Mock activities that are all status updates
        mock_activity1 = Mock()
        mock_activity1.is_status = True

        # First call - no previous state
        result1 = self.data._is_unchanged_push_state(
            self.device_id, message1, SOURCE_PUBNUB, [mock_activity1]
        )
        # Should not skip (need to process) and should track the state
        assert result1 is False

        # Verify state was tracked
        state_key = f"{self.device_id}:{SOURCE_PUBNUB}"
        assert state_key in self.data._last_push_state
        assert self.data._last_push_state[state_key] == {
            "lock": "locked",
            "door": "closed",
        }

    def test_pubnub_duplicate_status_update_skipped(self):
        """Test that duplicate PubNub status updates are skipped."""
        message = {
            "status": "locked",
            "doorState": "closed",
        }

        # Set initial state
        state_key = f"{self.device_id}:{SOURCE_PUBNUB}"
        self.data._last_push_state[state_key] = {
            "lock": "locked",
            "door": "closed",
        }

        # Mock activities that are all status updates
        mock_activity = Mock()
        mock_activity.is_status = True

        # Same state status update - should be skipped
        result = self.data._is_unchanged_push_state(
            self.device_id, message, SOURCE_PUBNUB, [mock_activity]
        )
        assert result is True  # Should skip

        # State should not have changed
        assert self.data._last_push_state[state_key] == {
            "lock": "locked",
            "door": "closed",
        }

    def test_pubnub_changed_status_update_not_tracked(self):
        """Test that changed PubNub status updates are processed but not tracked."""
        # Set initial state
        state_key = f"{self.device_id}:{SOURCE_PUBNUB}"
        self.data._last_push_state[state_key] = {
            "lock": "locked",
            "door": "closed",
        }

        # Different state in status update
        message = {
            "status": "unlocked",
            "doorState": "open",
        }

        # Mock activities that are all status updates
        mock_activity = Mock()
        mock_activity.is_status = True

        result = self.data._is_unchanged_push_state(
            self.device_id, message, SOURCE_PUBNUB, [mock_activity]
        )
        assert result is False  # Should process (state changed)

        # State should NOT have been updated (status updates don't track)
        assert self.data._last_push_state[state_key] == {
            "lock": "locked",
            "door": "closed",
        }

    def test_pubnub_real_action_updates_tracking(self):
        """Test that real PubNub actions update state tracking."""
        # Set initial state
        state_key = f"{self.device_id}:{SOURCE_PUBNUB}"
        self.data._last_push_state[state_key] = {
            "lock": "locked",
            "door": "closed",
        }

        # Real unlock action
        message = {
            "status": "unlocked",
            "doorState": "closed",
            "info": {"action": "unlock"},
            "callingUserID": "user123",
        }

        # Mock activity that is NOT a status update
        mock_activity = Mock()
        mock_activity.is_status = False

        result = self.data._is_unchanged_push_state(
            self.device_id, message, SOURCE_PUBNUB, [mock_activity]
        )
        assert result is False  # Should process

        # State SHOULD have been updated (real action)
        assert self.data._last_push_state[state_key] == {
            "lock": "unlocked",
            "door": "closed",
        }

    def test_status_update_between_real_actions_doesnt_interfere(self):
        """Test that status updates between real actions don't interfere with detection."""
        state_key = f"{self.device_id}:{SOURCE_PUBNUB}"

        # Step 1: Real unlock action
        message1 = {
            "status": "unlocked",
            "doorState": "closed",
        }
        mock_activity1 = Mock()
        mock_activity1.is_status = False

        result1 = self.data._is_unchanged_push_state(
            self.device_id, message1, SOURCE_PUBNUB, [mock_activity1]
        )
        assert result1 is False
        assert self.data._last_push_state[state_key] == {
            "lock": "unlocked",
            "door": "closed",
        }

        # Step 2: Status update with same state
        message2 = {
            "status": "unlocked",
            "doorState": "closed",
        }
        mock_activity2 = Mock()
        mock_activity2.is_status = True

        result2 = self.data._is_unchanged_push_state(
            self.device_id, message2, SOURCE_PUBNUB, [mock_activity2]
        )
        assert result2 is True  # Should skip (unchanged)
        assert self.data._last_push_state[state_key] == {
            "lock": "unlocked",
            "door": "closed",
        }  # State unchanged

        # Step 3: Real lock action
        message3 = {
            "status": "locked",
            "doorState": "closed",
        }
        mock_activity3 = Mock()
        mock_activity3.is_status = False

        result3 = self.data._is_unchanged_push_state(
            self.device_id, message3, SOURCE_PUBNUB, [mock_activity3]
        )
        assert result3 is False  # Should process (real action with changed state)
        assert self.data._last_push_state[state_key] == {
            "lock": "locked",
            "door": "closed",
        }  # State updated

    def test_websocket_always_tracks_state(self):
        """Test that WebSocket messages always track state changes."""
        state_key = f"{self.device_id}:{SOURCE_WEBSOCKET}"

        # First WebSocket message
        message1 = {
            "lockAction": "locked",
            "doorState": "closed",
        }

        result1 = self.data._is_unchanged_push_state(
            self.device_id, message1, SOURCE_WEBSOCKET, []
        )
        assert result1 is False
        assert self.data._last_push_state[state_key] == {
            "lock": "locked",
            "door": "closed",
        }

        # Same state - should skip
        result2 = self.data._is_unchanged_push_state(
            self.device_id, message1, SOURCE_WEBSOCKET, []
        )
        assert result2 is True

        # Different state - should process and track
        message2 = {
            "lockAction": "unlocked",
            "doorState": "open",
        }
        result3 = self.data._is_unchanged_push_state(
            self.device_id, message2, SOURCE_WEBSOCKET, []
        )
        assert result3 is False
        assert self.data._last_push_state[state_key] == {
            "lock": "unlocked",
            "door": "open",
        }

    def test_separate_tracking_per_source(self):
        """Test that state is tracked separately for each source."""
        pubnub_key = f"{self.device_id}:{SOURCE_PUBNUB}"
        websocket_key = f"{self.device_id}:{SOURCE_WEBSOCKET}"

        # Set PubNub state
        pubnub_message = {
            "status": "locked",
            "doorState": "closed",
        }
        mock_activity = Mock()
        mock_activity.is_status = False

        self.data._is_unchanged_push_state(
            self.device_id, pubnub_message, SOURCE_PUBNUB, [mock_activity]
        )

        # Set different WebSocket state
        websocket_message = {
            "lockAction": "unlocked",
            "doorState": "open",
        }
        self.data._is_unchanged_push_state(
            self.device_id, websocket_message, SOURCE_WEBSOCKET, []
        )

        # Verify states are tracked separately
        assert self.data._last_push_state[pubnub_key] == {
            "lock": "locked",
            "door": "closed",
        }
        assert self.data._last_push_state[websocket_key] == {
            "lock": "unlocked",
            "door": "open",
        }

    def test_unchanged_state_still_processes_newer_activities(self, caplog):
        """Test that unchanged state messages still process if they have newer activities."""

        # Create a more complete mock data object with required methods
        class TestDataWithMethods:
            def __init__(self):
                self._last_push_state = {}
                self._device_detail_by_id = {}
                self.activity_stream = Mock()
                self.activity_stream.async_process_newer_device_activities = Mock(
                    return_value=True
                )
                self.activity_stream.async_schedule_house_id_refresh = Mock()

            # Bind the actual methods from YaleXSData
            _is_unchanged_push_state = YaleXSData._is_unchanged_push_state
            _async_handle_push_message = YaleXSData._async_handle_push_message

            def get_device_detail(self, device_id):
                return self._device_detail_by_id.get(device_id)

            def async_signal_device_id_update(self, device_id):
                pass

        data = TestDataWithMethods()
        device_id = "test_device"

        # Create a mock device
        mock_device = Mock()
        mock_device.device_id = device_id
        mock_device.house_id = "test_house"
        data._device_detail_by_id[device_id] = mock_device

        # Set initial state
        state_key = f"{device_id}:{SOURCE_PUBNUB}"
        data._last_push_state[state_key] = {
            "lock": "locked",
            "door": "closed",
        }

        # Message with same state but newer timestamp (would have newer activities)
        message = {
            "status": "locked",
            "doorState": "closed",
            "callingUserID": "manuallock",
        }

        # Mock activity that is not a status update
        mock_activity = Mock()
        mock_activity.is_status = False
        mock_activity.action = "lock"

        with (
            patch(
                "yalexs.manager.data.activities_from_pubnub_message"
            ) as mock_activities_func,
            caplog.at_level(logging.DEBUG),
        ):
            mock_activities_func.return_value = [mock_activity]

            # Call the push message handler
            data._async_handle_push_message(
                device_id, datetime.now(), message, SOURCE_PUBNUB
            )

            # Verify activities were processed even though state unchanged
            assert data.activity_stream.async_process_newer_device_activities.called
            assert data.activity_stream.async_process_newer_device_activities.call_args[
                0
            ][0] == [mock_activity]

            # Verify we logged that state was unchanged
            assert any(
                "Skipping unchanged" in record.message for record in caplog.records
            )

            # Verify house refresh was NOT scheduled (because unchanged)
            assert not data.activity_stream.async_schedule_house_id_refresh.called


@pytest.mark.asyncio
async def test_fetch_lock_capabilities() -> None:
    """Test that lock capabilities are fetched and set correctly."""
    # Create mock gateway and API
    mock_gateway = Mock()
    mock_gateway.async_get_access_token = AsyncMock(return_value="test-token")

    mock_api = Mock()
    mock_gateway.api = mock_api
    mock_gateway.api.brand = Brand.YALE_HOME  # Set brand for capability fetching

    # Create MockYaleXSData instance
    data = MockYaleXSData(mock_gateway)

    # Create mock lock details
    lock_detail_1 = Mock(spec=LockDetail)
    lock_detail_1.device_name = "Front Door"
    lock_detail_1.set_capabilities = Mock()

    lock_detail_2 = Mock(spec=LockDetail)
    lock_detail_2.device_name = "Back Door"
    lock_detail_2.set_capabilities = Mock()

    # Set up device details
    # Note: lock_id is the serial number for locks
    data._device_detail_by_id = {
        "SERIAL1": lock_detail_1,
        "SERIAL2": lock_detail_2,
        "doorbell1": Mock(),  # Not a lock, should be skipped
    }
    data._locks_by_id = {
        "SERIAL1": Mock(),
        "SERIAL2": Mock(),
    }

    # Mock API responses
    capabilities_1: CapabilitiesResponse = {
        "lock": {
            "unlatch": True,
            "doorSense": True,
            "batteryType": "AA",
        }
    }
    capabilities_2: CapabilitiesResponse = {
        "lock": {
            "unlatch": False,
            "doorSense": False,
            "batteryType": "CR123",
        }
    }

    # Configure mock API
    async def mock_get_capabilities(token: str, serial: str) -> CapabilitiesResponse:
        if serial == "SERIAL1":
            return capabilities_1
        if serial == "SERIAL2":
            return capabilities_2
        raise ValueError(f"Unknown serial: {serial}")

    mock_api.async_get_lock_capabilities = AsyncMock(side_effect=mock_get_capabilities)

    # Call the method
    await data._async_fetch_lock_capabilities()

    # Verify API was called with correct parameters
    assert mock_api.async_get_lock_capabilities.call_count == 2
    mock_api.async_get_lock_capabilities.assert_any_call("test-token", "SERIAL1")
    mock_api.async_get_lock_capabilities.assert_any_call("test-token", "SERIAL2")

    # Verify capabilities were set on lock details
    lock_detail_1.set_capabilities.assert_called_once_with(capabilities_1)
    lock_detail_2.set_capabilities.assert_called_once_with(capabilities_2)


@pytest.mark.asyncio
async def test_fetch_lock_capabilities_with_error(
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test that capability fetch errors are handled gracefully."""
    # Create mock gateway and API
    mock_gateway = Mock()
    mock_gateway.async_get_access_token = AsyncMock(return_value="test-token")

    mock_api = Mock()
    mock_gateway.api = mock_api
    mock_gateway.api.brand = Brand.YALE_HOME  # Set brand for capability fetching

    # Create MockYaleXSData instance
    data = MockYaleXSData(mock_gateway)

    # Create mock lock detail
    lock_detail = Mock(spec=LockDetail)
    lock_detail.device_name = "Front Door"
    lock_detail.set_capabilities = Mock()

    # Set up device details (lock_id is serial number)
    data._device_detail_by_id = {
        "SERIAL1": lock_detail,
    }
    data._locks_by_id = {
        "SERIAL1": Mock(),
    }

    # Mock API to raise an error

    mock_api.async_get_lock_capabilities = AsyncMock(
        side_effect=ClientError("API Error")
    )

    # Call the method with logging
    with caplog.at_level(logging.WARNING):
        await data._async_fetch_lock_capabilities()

    # Verify API was called
    mock_api.async_get_lock_capabilities.assert_called_once_with(
        "test-token", "SERIAL1"
    )

    # Verify capabilities were NOT set due to error
    lock_detail.set_capabilities.assert_not_called()

    # Verify error was logged
    assert "Failed to fetch capabilities for lock Front Door" in caplog.text
    assert "API Error" in caplog.text


@pytest.mark.asyncio
async def test_fetch_lock_capabilities_skips_non_locks() -> None:
    """Test that non-lock devices are skipped when fetching capabilities."""
    # Create mock gateway and API
    mock_gateway = Mock()
    mock_gateway.async_get_access_token = AsyncMock(return_value="test-token")

    mock_api = Mock()
    mock_gateway.api = mock_api
    mock_gateway.api.brand = Brand.YALE_HOME  # Set brand for capability fetching

    # Create MockYaleXSData instance
    data = MockYaleXSData(mock_gateway)

    # Create mock lock detail
    lock_detail = Mock(spec=LockDetail)
    lock_detail.device_name = "Front Door"
    lock_detail.set_capabilities = Mock()

    # Set up device details with mixed devices
    data._device_detail_by_id = {
        "SERIAL1": lock_detail,  # This is a lock
        "doorbell1": Mock(),  # This is not a lock
        "doorbell2": Mock(),  # This is not a lock
    }
    data._locks_by_id = {
        "SERIAL1": Mock(),  # Only this one is a lock
    }

    # Mock API response
    capabilities: CapabilitiesResponse = {
        "lock": {
            "unlatch": True,
            "doorSense": True,
            "batteryType": "AA",
        }
    }

    mock_api.async_get_lock_capabilities = AsyncMock(return_value=capabilities)

    # Call the method
    await data._async_fetch_lock_capabilities()

    # Verify API was called only once (for the lock, not the doorbells)
    mock_api.async_get_lock_capabilities.assert_called_once_with(
        "test-token", "SERIAL1"
    )

    # Verify capabilities were set only on the lock
    lock_detail.set_capabilities.assert_called_once_with(capabilities)


@pytest.mark.asyncio
async def test_fetch_lock_capabilities_sequential_execution() -> None:
    """Test that capabilities are fetched sequentially, not in parallel."""
    # Create mock gateway and API
    mock_gateway = Mock()
    mock_gateway.async_get_access_token = AsyncMock(return_value="test-token")

    mock_api = Mock()
    mock_gateway.api = mock_api
    mock_gateway.api.brand = Brand.YALE_HOME  # Set brand for capability fetching

    # Create MockYaleXSData instance
    data = MockYaleXSData(mock_gateway)

    # Create mock lock details
    lock_detail_1 = Mock(spec=LockDetail)
    lock_detail_1.device_name = "Front Door"
    lock_detail_1.set_capabilities = Mock()

    lock_detail_2 = Mock(spec=LockDetail)
    lock_detail_2.device_name = "Back Door"
    lock_detail_2.set_capabilities = Mock()

    lock_detail_3 = Mock(spec=LockDetail)
    lock_detail_3.device_name = "Side Door"
    lock_detail_3.set_capabilities = Mock()

    # Set up device details
    data._device_detail_by_id = {
        "SERIAL1": lock_detail_1,
        "SERIAL2": lock_detail_2,
        "SERIAL3": lock_detail_3,
    }
    data._locks_by_id = {
        "SERIAL1": Mock(),
        "SERIAL2": Mock(),
        "SERIAL3": Mock(),
    }

    # Track call order
    call_order: list[str] = []

    async def mock_get_capabilities(token: str, serial: str) -> CapabilitiesResponse:
        call_order.append(serial)
        return {"lock": {"unlatch": True}}

    mock_api.async_get_lock_capabilities = AsyncMock(side_effect=mock_get_capabilities)

    # Call the method
    await data._async_fetch_lock_capabilities()

    # Verify all were called in sequence
    assert call_order == ["SERIAL1", "SERIAL2", "SERIAL3"]
    assert mock_api.async_get_lock_capabilities.call_count == 3


@pytest.mark.asyncio
async def test_august_brand_does_not_fetch_capabilities():
    """Test that August brand does not fetch device capabilities."""
    # Create mock gateway with August brand
    mock_gateway = AsyncMock()
    mock_gateway.brand = Brand.AUGUST  # August brand
    mock_gateway.access_token = "test-token"
    mock_gateway.async_get_access_token = AsyncMock(return_value="test-token")

    mock_api = Mock()
    mock_gateway.api = mock_api
    mock_gateway.api.brand = Brand.AUGUST  # Set August brand for API

    # Create MockYaleXSData instance
    data = MockYaleXSData(mock_gateway)

    # Set up test locks
    lock1 = {
        "LockID": "ABC1",
        "LockName": "Lock 1",
        "HouseID": "house1",
        "SerialNumber": "SERIAL1",
        "Type": 5,
        "battery": 0.8,
        "currentFirmwareVersion": "1.0.0",
        "LockStatus": {"status": "locked"},
    }
    lock2 = {
        "LockID": "ABC2",
        "LockName": "Lock 2",
        "HouseID": "house1",
        "SerialNumber": "SERIAL2",
        "Type": 17,
        "battery": 0.9,
        "currentFirmwareVersion": "1.0.0",
        "LockStatus": {"status": "unlocked"},
    }

    data._lock_details = {
        "ABC1": LockDetail(lock1),
        "ABC2": LockDetail(lock2),
    }

    # Mock the capabilities fetch - should not be called
    mock_api.async_get_lock_capabilities = AsyncMock()

    # Call the method
    await data._async_fetch_lock_capabilities()

    # Verify the capabilities method was NOT called for August brand
    assert mock_api.async_get_lock_capabilities.call_count == 0


@pytest.mark.asyncio
async def test_fetch_lock_capabilities_handles_404_and_409_errors(
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test that 404 and 409 errors are handled gracefully when fetching capabilities."""
    # Create mock gateway and API
    mock_gateway = Mock()
    mock_gateway.async_get_access_token = AsyncMock(return_value="test-token")

    mock_api = Mock()
    mock_gateway.api = mock_api
    mock_gateway.api.brand = Brand.YALE_HOME  # Set brand for capability fetching

    # Create MockYaleXSData instance
    data = MockYaleXSData(mock_gateway)

    # Create mock lock details
    lock_detail_404 = Mock(spec=LockDetail)
    lock_detail_404.device_name = "Lock 404"
    lock_detail_404.set_capabilities = Mock()

    lock_detail_409 = Mock(spec=LockDetail)
    lock_detail_409.device_name = "Lock 409"
    lock_detail_409.set_capabilities = Mock()

    # Set up device details
    data._device_detail_by_id = {
        "SERIAL404": lock_detail_404,
        "SERIAL409": lock_detail_409,
    }
    data._locks_by_id = {
        "SERIAL404": Mock(),
        "SERIAL409": Mock(),
    }

    # Mock API to raise 404 and 409 errors
    async def mock_get_capabilities(token: str, serial: str) -> None:
        if serial == "SERIAL404":
            error = YaleApiError(
                "The operation failed with error code 404: Device info not found.",
                ClientResponseError(
                    request_info=Mock(),
                    history=(),
                    status=404,
                    message="Device info not found",
                ),
            )
            raise error
        if serial == "SERIAL409":
            error = YaleApiError(
                "The operation failed with error code 409: Cannot infer deviceType from serial number.",
                ClientResponseError(
                    request_info=Mock(),
                    history=(),
                    status=409,
                    message="Cannot infer deviceType from serial number.",
                ),
            )
            raise error

    mock_api.async_get_lock_capabilities = AsyncMock(side_effect=mock_get_capabilities)

    # Call the method with debug logging
    with caplog.at_level(logging.DEBUG):
        await data._async_fetch_lock_capabilities()

    # Verify API was called for both locks
    assert mock_api.async_get_lock_capabilities.call_count == 2

    # Verify capabilities were NOT set due to errors
    lock_detail_404.set_capabilities.assert_not_called()
    lock_detail_409.set_capabilities.assert_not_called()

    # Verify debug messages were logged (not warnings)
    assert "Cannot fetch capabilities for lock Lock 404 (HTTP 404)" in caplog.text
    assert "Cannot fetch capabilities for lock Lock 409 (HTTP 409)" in caplog.text
    # Verify no warning logs for these expected errors
    warning_records = [r for r in caplog.records if r.levelno >= logging.WARNING]
    assert len(warning_records) == 0


@pytest.mark.asyncio
async def test_fetch_lock_capabilities_handles_other_errors_with_warning(
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test that non-404/409 errors are logged as warnings when fetching capabilities."""
    # Create mock gateway and API
    mock_gateway = Mock()
    mock_gateway.async_get_access_token = AsyncMock(return_value="test-token")

    mock_api = Mock()
    mock_gateway.api = mock_api
    mock_gateway.api.brand = Brand.YALE_HOME  # Set brand for capability fetching

    # Create MockYaleXSData instance
    data = MockYaleXSData(mock_gateway)

    # Create mock lock details
    lock_detail_401 = Mock(spec=LockDetail)
    lock_detail_401.device_name = "Lock 401"
    lock_detail_401.set_capabilities = Mock()

    lock_detail_500 = Mock(spec=LockDetail)
    lock_detail_500.device_name = "Lock 500"
    lock_detail_500.set_capabilities = Mock()

    # Set up device details
    data._device_detail_by_id = {
        "SERIAL401": lock_detail_401,
        "SERIAL500": lock_detail_500,
    }
    data._locks_by_id = {
        "SERIAL401": Mock(),
        "SERIAL500": Mock(),
    }

    # Mock API to raise 401 and 500 errors
    async def mock_get_capabilities(token: str, serial: str) -> None:
        if serial == "SERIAL401":
            error = YaleApiError(
                "The operation failed with error code 401: Unauthorized.",
                ClientResponseError(
                    request_info=Mock(),
                    history=(),
                    status=401,
                    message="Unauthorized",
                ),
            )
            raise error
        if serial == "SERIAL500":
            error = YaleApiError(
                "The operation failed with error code 500: Internal Server Error.",
                ClientResponseError(
                    request_info=Mock(),
                    history=(),
                    status=500,
                    message="Internal Server Error",
                ),
            )
            raise error

    mock_api.async_get_lock_capabilities = AsyncMock(side_effect=mock_get_capabilities)

    # Call the method with warning logging
    with caplog.at_level(logging.WARNING):
        await data._async_fetch_lock_capabilities()

    # Verify API was called for both locks
    assert mock_api.async_get_lock_capabilities.call_count == 2

    # Verify capabilities were NOT set due to errors
    lock_detail_401.set_capabilities.assert_not_called()
    lock_detail_500.set_capabilities.assert_not_called()

    # Verify warning messages were logged for non-404/409 errors
    assert "Failed to fetch capabilities for lock Lock 401 (HTTP 401)" in caplog.text
    assert "Failed to fetch capabilities for lock Lock 500 (HTTP 500)" in caplog.text

    # Verify these are logged as warnings, not debug
    warning_records = [r for r in caplog.records if r.levelno >= logging.WARNING]
    assert len(warning_records) == 2


@pytest.mark.asyncio
async def test_fetch_lock_capabilities_handles_network_errors(
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test that network errors like ClientError and TimeoutError are handled gracefully."""
    # Create mock gateway and API
    mock_gateway = Mock()
    mock_gateway.async_get_access_token = AsyncMock(return_value="test-token")

    mock_api = Mock()
    mock_gateway.api = mock_api
    mock_gateway.api.brand = Brand.YALE_HOME  # Set brand for capability fetching

    # Create MockYaleXSData instance
    data = MockYaleXSData(mock_gateway)

    # Create mock lock details
    lock_detail_timeout = Mock(spec=LockDetail)
    lock_detail_timeout.device_name = "Lock Timeout"
    lock_detail_timeout.set_capabilities = Mock()

    lock_detail_network = Mock(spec=LockDetail)
    lock_detail_network.device_name = "Lock Network"
    lock_detail_network.set_capabilities = Mock()

    # Set up device details
    data._device_detail_by_id = {
        "SERIAL_TIMEOUT": lock_detail_timeout,
        "SERIAL_NETWORK": lock_detail_network,
    }
    data._locks_by_id = {
        "SERIAL_TIMEOUT": Mock(),
        "SERIAL_NETWORK": Mock(),
    }

    # Mock API to raise TimeoutError and ClientError
    async def mock_get_capabilities(token: str, serial: str) -> None:
        if serial == "SERIAL_TIMEOUT":
            raise TimeoutError("Request timed out")
        if serial == "SERIAL_NETWORK":
            raise ClientError("Network error")

    mock_api.async_get_lock_capabilities = AsyncMock(side_effect=mock_get_capabilities)

    # Call the method with warning logging
    with caplog.at_level(logging.WARNING):
        await data._async_fetch_lock_capabilities()

    # Verify API was called for both locks
    assert mock_api.async_get_lock_capabilities.call_count == 2

    # Verify capabilities were NOT set due to errors
    lock_detail_timeout.set_capabilities.assert_not_called()
    lock_detail_network.set_capabilities.assert_not_called()

    # Verify warning messages were logged for network errors
    assert (
        "Failed to fetch capabilities for lock Lock Timeout: Request timed out"
        in caplog.text
    )
    assert (
        "Failed to fetch capabilities for lock Lock Network: Network error"
        in caplog.text
    )

    # Verify these are logged as warnings
    warning_records = [r for r in caplog.records if r.levelno >= logging.WARNING]
    assert len(warning_records) == 2


@pytest.mark.asyncio
async def test_async_operate_lock_wait_mode() -> None:
    """Test async_operate_lock LOCK operation when waiting for response."""
    mock_gateway = Mock()
    mock_gateway.async_get_access_token = AsyncMock(return_value="test_token")
    mock_api = Mock()

    data = MockYaleXSData(mock_gateway, mock_api)

    # Mock the individual lock operation methods
    data.async_lock = AsyncMock(return_value=["lock_activity"])
    data.async_lock_async = AsyncMock(return_value="lock_request_id")

    device_id = "test_device"

    # Mock device detail without unlatch support
    mock_detail = Mock(spec=LockDetail)
    mock_detail.unlatch_supported = False
    data.get_device_detail = Mock(return_value=mock_detail)

    result = await data.async_operate_lock(
        device_id, LockOperation.LOCK, push_updates_connected=False
    )

    assert result == ["lock_activity"]
    data.async_lock.assert_called_once_with(device_id)
    data.async_lock_async.assert_not_called()


@pytest.mark.asyncio
async def test_async_operate_lock_push_mode() -> None:
    """Test async_operate_lock LOCK operation with push updates (no wait)."""
    mock_gateway = Mock()
    mock_gateway.async_get_access_token = AsyncMock(return_value="test_token")
    mock_api = Mock()

    data = MockYaleXSData(mock_gateway, mock_api)

    # Mock the individual lock operation methods
    data.async_lock = AsyncMock(return_value=["lock_activity"])
    data.async_lock_async = AsyncMock(return_value="lock_request_id")

    device_id = "test_device"

    # Mock device detail without unlatch support
    mock_detail = Mock(spec=LockDetail)
    mock_detail.unlatch_supported = False
    data.get_device_detail = Mock(return_value=mock_detail)

    result = await data.async_operate_lock(
        device_id, LockOperation.LOCK, push_updates_connected=True, hyper_bridge=True
    )

    assert result == []  # Returns empty list when not waiting
    data.async_lock.assert_not_called()
    data.async_lock_async.assert_called_once_with(device_id, True)


@pytest.mark.asyncio
async def test_async_operate_unlock_no_unlatch_support() -> None:
    """Test async_operate_lock UNLOCK operation when device doesn't support unlatch."""
    mock_gateway = Mock()
    mock_gateway.async_get_access_token = AsyncMock(return_value="test_token")
    mock_api = Mock()

    data = MockYaleXSData(mock_gateway, mock_api)

    # Mock the individual lock operation methods
    data.async_unlock = AsyncMock(return_value=["unlock_activity"])
    data.async_unlatch = AsyncMock(return_value=["unlatch_activity"])

    device_id = "test_device"

    # Mock device detail without unlatch support
    mock_detail = Mock(spec=LockDetail)
    mock_detail.unlatch_supported = False
    data.get_device_detail = Mock(return_value=mock_detail)

    result = await data.async_operate_lock(
        device_id, LockOperation.UNLOCK, push_updates_connected=False
    )

    assert result == ["unlock_activity"]
    data.async_unlock.assert_called_once_with(device_id)
    data.async_unlatch.assert_not_called()


@pytest.mark.asyncio
async def test_async_operate_unlock_with_unlatch_support() -> None:
    """Test async_operate_lock UNLOCK operation when device supports unlatch (should call unlatch)."""
    mock_gateway = Mock()
    mock_gateway.async_get_access_token = AsyncMock(return_value="test_token")
    mock_api = Mock()

    data = MockYaleXSData(mock_gateway, mock_api)

    # Mock the individual lock operation methods
    data.async_unlock = AsyncMock(return_value=["unlock_activity"])
    data.async_unlatch = AsyncMock(return_value=["unlatch_activity"])

    device_id = "test_device"

    # Mock device detail WITH unlatch support
    mock_detail = Mock(spec=LockDetail)
    mock_detail.unlatch_supported = True
    data.get_device_detail = Mock(return_value=mock_detail)

    result = await data.async_operate_lock(
        device_id, LockOperation.UNLOCK, push_updates_connected=False
    )

    # When unlatch is supported, UNLOCK should call unlatch!
    assert result == ["unlatch_activity"]
    data.async_unlatch.assert_called_once_with(device_id)
    data.async_unlock.assert_not_called()


@pytest.mark.asyncio
async def test_async_operate_open_no_unlatch_support() -> None:
    """Test async_operate_lock OPEN operation when device doesn't support unlatch."""
    mock_gateway = Mock()
    mock_gateway.async_get_access_token = AsyncMock(return_value="test_token")
    mock_api = Mock()

    data = MockYaleXSData(mock_gateway, mock_api)

    # Mock the individual lock operation methods
    data.async_unlock = AsyncMock(return_value=["unlock_activity"])
    data.async_unlatch = AsyncMock(return_value=["unlatch_activity"])

    device_id = "test_device"

    # Mock device detail without unlatch support
    mock_detail = Mock(spec=LockDetail)
    mock_detail.unlatch_supported = False
    data.get_device_detail = Mock(return_value=mock_detail)

    result = await data.async_operate_lock(
        device_id, LockOperation.OPEN, push_updates_connected=False
    )

    assert result == ["unlatch_activity"]
    data.async_unlatch.assert_called_once_with(device_id)
    data.async_unlock.assert_not_called()


@pytest.mark.asyncio
async def test_async_operate_open_with_unlatch_support() -> None:
    """Test async_operate_lock OPEN operation when device supports unlatch (should call unlock)."""
    mock_gateway = Mock()
    mock_gateway.async_get_access_token = AsyncMock(return_value="test_token")
    mock_api = Mock()

    data = MockYaleXSData(mock_gateway, mock_api)

    # Mock the individual lock operation methods
    data.async_unlock = AsyncMock(return_value=["unlock_activity"])
    data.async_unlatch = AsyncMock(return_value=["unlatch_activity"])

    device_id = "test_device"

    # Mock device detail WITH unlatch support
    mock_detail = Mock(spec=LockDetail)
    mock_detail.unlatch_supported = True
    data.get_device_detail = Mock(return_value=mock_detail)

    result = await data.async_operate_lock(
        device_id, LockOperation.OPEN, push_updates_connected=False
    )

    # When unlatch is supported, OPEN should call unlock!
    assert result == ["unlock_activity"]
    data.async_unlock.assert_called_once_with(device_id)
    data.async_unlatch.assert_not_called()


@pytest.mark.asyncio
async def test_async_operate_lock_all_operations_with_push() -> None:
    """Test async_operate_lock all operations with push updates enabled."""
    mock_gateway = Mock()
    mock_gateway.async_get_access_token = AsyncMock(return_value="test_token")
    mock_api = Mock()

    data = MockYaleXSData(mock_gateway, mock_api)

    # Mock all async operation methods
    data.async_lock_async = AsyncMock(return_value="lock_request_id")
    data.async_unlock_async = AsyncMock(return_value="unlock_request_id")
    data.async_unlatch_async = AsyncMock(return_value="unlatch_request_id")

    device_id = "test_device"

    # Mock device detail with unlatch support
    mock_detail = Mock(spec=LockDetail)
    mock_detail.unlatch_supported = True
    data.get_device_detail = Mock(return_value=mock_detail)

    # Test LOCK operation
    result = await data.async_operate_lock(
        device_id, LockOperation.LOCK, push_updates_connected=True
    )
    assert result == []
    data.async_lock_async.assert_called_once_with(device_id, True)

    # Reset mocks
    data.async_lock_async.reset_mock()
    data.async_unlock_async.reset_mock()
    data.async_unlatch_async.reset_mock()

    # Test UNLOCK (should call unlatch_async when unlatch supported)
    result = await data.async_operate_lock(
        device_id, LockOperation.UNLOCK, push_updates_connected=True
    )
    assert result == []
    data.async_unlatch_async.assert_called_once_with(device_id, True)

    # Reset mocks
    data.async_lock_async.reset_mock()
    data.async_unlock_async.reset_mock()
    data.async_unlatch_async.reset_mock()

    # Test OPEN (should call unlock_async when unlatch supported)
    result = await data.async_operate_lock(
        device_id, LockOperation.OPEN, push_updates_connected=True
    )
    assert result == []
    data.async_unlock_async.assert_called_once_with(device_id, True)


@pytest.mark.asyncio
async def test_async_operate_lock_invalid_operation() -> None:
    """Test async_operate_lock with invalid operation raises ValueError."""
    mock_gateway = Mock()
    mock_gateway.async_get_access_token = AsyncMock(return_value="test_token")
    mock_api = Mock()

    data = MockYaleXSData(mock_gateway, mock_api)

    device_id = "test_device"

    # Mock device detail
    mock_detail = Mock(spec=LockDetail)
    mock_detail.unlatch_supported = False
    data.get_device_detail = Mock(return_value=mock_detail)

    # Create an invalid operation
    invalid_op: Any = Mock()
    invalid_op.value = "invalid"

    with pytest.raises(ValueError, match="Invalid operation"):
        await data.async_operate_lock(
            device_id, invalid_op, push_updates_connected=False
        )


@pytest.mark.asyncio
async def test_async_operate_lock_no_device_detail() -> None:
    """Test async_operate_lock when get_device_detail returns None."""
    mock_gateway = Mock()
    mock_gateway.async_get_access_token = AsyncMock(return_value="test_token")
    mock_api = Mock()

    data = MockYaleXSData(mock_gateway, mock_api)

    # Mock the individual lock operation methods
    data.async_unlock = AsyncMock(return_value=["unlock_activity"])
    data.async_unlatch = AsyncMock(return_value=["unlatch_activity"])

    device_id = "test_device"

    # Mock get_device_detail to return None
    data.get_device_detail = Mock(return_value=None)

    result = await data.async_operate_lock(
        device_id, LockOperation.UNLOCK, push_updates_connected=False
    )

    # Should use normal mapping (unlock -> unlock)
    assert result == ["unlock_activity"]
    data.async_unlock.assert_called_once_with(device_id)
    data.async_unlatch.assert_not_called()



from __future__ import annotations

from typing import Any

from ._compat import cached_property


class Device:
    """Base class for all devices."""

    def __init__(self, device_id: str, device_name: str, house_id: str) -> None:
        self._device_id = device_id
        self._device_name = device_name
        self._house_id = house_id

    @cached_property
    def device_id(self) -> str:
        return self._device_id

    @cached_property
    def device_name(self) -> str:
        return self._device_name

    @cached_property
    def house_id(self) -> str:
        return self._house_id


class DeviceDetail:
    def __init__(
        self,
        device_id: str,
        device_name: str,
        house_id: str,
        serial_number: str,
        firmware_version: str,
        pubsub_channel: str,
        data: dict[str, Any],
    ) -> None:
        self._device_id = device_id
        self._device_name = device_name
        self._house_id = house_id
        self._serial_number = serial_number
        self._firmware_version = firmware_version
        self._pubsub_channel = pubsub_channel
        self._data = data

    @cached_property
    def raw(self):
        return self._data

    @cached_property
    def device_id(self):
        return self._device_id

    @cached_property
    def device_name(self):
        return self._device_name

    @cached_property
    def house_id(self):
        return self._house_id

    @cached_property
    def serial_number(self):
        return self._serial_number

    @cached_property
    def firmware_version(self):
        return self._firmware_version

    @cached_property
    def pubsub_channel(self):
        return self._pubsub_channel
